import os
import sys
import logging

from jet_bridge_base.logger import set_logger_level

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
CURRENT_MODULE = sys.modules[__name__]
DEFAULT_CONFIG_PATH = os.path.join('jet.conf')

DEBUG = False
READ_ONLY = False
AUTO_OPEN_REGISTER = True
CONFIG = DEFAULT_CONFIG_PATH
PROJECT = None
TOKEN = None
ENVIRONMENT = None
CORS_HEADERS = True
BASE_URL = None
JWT_VERIFY_KEY = '-----BEGIN PUBLIC KEY-----\nMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAyfJablscZmsh7eHisWHi\n/x2gJUjc9juL4VLkEpp6PH9Ah+wBtGmPF9nfoVlQsgH9hra63POiaLUzIwW/ewOl\njyPF0FJngsxreCxNs8lmd/WNXrcnazknFItkFFeXJuMTfoBPQGiZOmVYb14jmvkc\n9vMmeXYjrbT+95SVayN3E6DzLoHDhny4Mka1OsxvIP5s77s0dOo68TzoEfBVeuto\nI/dopG86DVu4wYVtYPITzJ4z47OFVPKCyYVyy5aR3+DUnmdK7xTRVr+iWmHpcr7e\nhoeVcL4CqAILZ0gd54kQmnHbg7Bu6x8JtQkiLU5TQvWzjiN00io4eydvIAkQTAaR\nmdd32O1vJbSHmLyCR2tEW/uV7P25naPUlkApxuLzh5C21S0XJxNJ/P07KSMymt5U\n1lWqt4CInpjAwMI8qs9MkEwJev5+yumxqIrDKcQLMR3TBLJZIb+rL1teCLOW28qB\nL6VSKhfKRIaXUdLpRwAcSuXraTzwa9oCCZa19tw3uizMeMFrCrv43YbyOsS9h7JQ\n8ixj/a1R/ud0fCrhXWUl7nKlz0b15koILLG1Ts+MUTmIaEnHTVEY74CfJVq7waw9\nx2kyzSzbsmMXvFkrVzTmyImTN631+gatU+npJ3vtcD9SooEZLOCLa4pb+DIsv9P1\nEeIEAh1VZC7s2qsQZsiYTG0CAwEAAQ==\n-----END PUBLIC KEY-----\n'
BEARER_AUTH_KEY = None
ENVIRONMENT_TYPE = None
BLACKLIST_HOSTS = None

WEB_BASE_URL = None
API_BASE_URL = None

DATABASE_ENGINE = None
DATABASE_URL = None
DATABASE_HOST = None
DATABASE_PORT = None
DATABASE_USER = None
DATABASE_PASSWORD = None
DATABASE_NAME = None
DATABASE_EXTRA = None
DATABASE_CONNECTIONS = None
DATABASE_CONNECTIONS_OVERFLOW = None
DATABASE_ONLY = None
DATABASE_EXCEPT = None
DATABASE_MAX_TABLES = None
DATABASE_SCHEMA = None
DATABASE_REFLECT_MAX_RECORDS = None

DATABASE_SSL_CA = None
DATABASE_SSL_CERT = None
DATABASE_SSL_KEY = None

DATABASE_SSH_HOST = None
DATABASE_SSH_PORT = None
DATABASE_SSH_USER = None
DATABASE_SSH_PRIVATE_KEY = None

COOKIE_SAMESITE = None
COOKIE_SECURE = None
COOKIE_DOMAIN = None
COOKIE_COMPRESS = None

STORE_PATH = None

CACHE_METADATA = False
CACHE_METADATA_PATH = None
CACHE_MODEL_DESCRIPTIONS = False

SSO_APPLICATIONS = {}

ALLOW_ORIGIN = '*'

TRACK_DATABASES = ''
TRACK_DATABASES_ENDPOINT = ''
TRACK_DATABASES_AUTH = ''

TRACK_MODELS_ENDPOINT = ''
TRACK_MODELS_AUTH = ''

TRACK_QUERY_SLOW_TIME = None
TRACK_QUERY_HIGH_MEMORY = None

RELEASE_INACTIVE_GRAPHQL_SCHEMAS_TIMEOUT = None

DISABLE_AUTH = None


def set_settings(settings):
    for key, value in settings.items():
        if value is None:
            continue
        setattr(CURRENT_MODULE, key, value)

    level = logging.DEBUG if DEBUG else logging.INFO
    set_logger_level(level)



from .mongo_mapper import MongoMapper
from .mongo_record import MongoRecord


class MongoDeclarativeMeta(object):
    def __init__(self, table):
        self._mapper = MongoMapper(table)

        for column in table.columns:
            setattr(self, column.name, column)

    def get_name(self):
        return self._mapper.selectable.name

    def __call__(self, **kwargs):
        table_name = self._mapper.selectable.name
        return MongoRecord(table_name, **kwargs)

    def __repr__(self):
        return self.get_name()



from jet_bridge_base.responses.base import Response
from jet_bridge_base.status import HTTP_302_FOUND


class RedirectResponse(Response):

    def __init__(self, url, status=HTTP_302_FOUND):
        self.url = url
        super(RedirectResponse, self).__init__(status=status)



import re
from sqlalchemy.exc import SQLAlchemyError

from .common import sql_get_session_engine


def fetch_postgresql_type_code_to_sql_type(session):
    from sqlalchemy.dialects.postgresql.base import ischema_names

    result = {}

    try:
        types_cursor = session.execute('''
            SELECT 
                pg_catalog.format_type(oid, NULL),
                oid 
            FROM 
                pg_type
        ''')

        for pg_type in types_cursor:
            # Copied from:
            # def _get_column_info
            # site-packages/sqlalchemy/dialects/postgresql/base.py:3716

            type_code = pg_type['oid']
            format_type = pg_type['format_type']

            def _handle_array_type(attype):
                return (
                    # strip '[]' from integer[], etc.
                    re.sub(r"\[\]$", "", attype),
                    attype.endswith("[]"),
                )

            # strip (*) from character varying(5), timestamp(5)
            # with time zone, geometry(POLYGON), etc.
            attype = re.sub(r"\(.*\)", "", format_type)

            # strip '[]' from integer[], etc. and check if an array
            attype, is_array = _handle_array_type(attype)

            if attype.startswith('interval'):
                attype = 'interval'

            sql_type = ischema_names.get(attype)

            if sql_type:
                result[type_code] = sql_type

        session.commit()
    except SQLAlchemyError:
        session.rollback()

    return result


def fetch_type_code_to_sql_type(session):
    if sql_get_session_engine(session) == 'postgresql':
        return fetch_postgresql_type_code_to_sql_type(session)



from jet_bridge_base.permissions import AdministratorPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.views.base.api import BaseAPIView


class TriggerExceptionView(BaseAPIView):
    permission_classes = (AdministratorPermissions,)

    def get(self, request, *args, **kwargs):
        division_by_zero = 1 / 0
        return JSONResponse({
            'result': True
        })



from datetime import tzinfo, timedelta


try:
    from datetime import timezone
    utc = timezone.utc
except ImportError:
    # Python 2
    class UTC(tzinfo):
        def utcoffset(self, dt):
            return timedelta(0)

        def tzname(self, dt):
            return "UTC"

        def dst(self, dt):
            return timedelta(0)

    utc = UTC()



import os

from jet_bridge_base.configuration import configuration
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.views.base.api import APIView


class FileUploadView(APIView):
    permission_classes = (HasProjectPermissions,)

    def post(self, request, *args, **kwargs):
        # TODO: Move to serializer
        original_filename, file = request.files.get('file', None)
        path = request.get_body_argument('path')
        filename = request.get_body_argument('filename', original_filename)

        upload_path = os.path.join(path, filename)
        upload_path = configuration.media_get_available_name(upload_path)

        # TODO: Add validations

        configuration.media_save(upload_path, file)

        return JSONResponse({
            'uploaded_path': upload_path,
            'uploaded_url': configuration.media_url(upload_path, request)
        })



from jet_bridge_base.db_types import inspect_uniform
from jet_bridge_base.serializers.model_serializer import ModelSerializer


def get_model_serializer(Model):
    mapper = inspect_uniform(Model)

    class CustomModelSerializer(ModelSerializer):
        class Meta:
            model = Model
            model_fields = list(map(lambda x: x.key, mapper.columns))

    return CustomModelSerializer



from jet_bridge_base import status
from jet_bridge_base.exceptions.api import APIException


class SqlError(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'Query failed'
    default_code = 'query_failed'



from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.exceptions.not_found import NotFound
from jet_bridge_base.external_auth.utils import load_strategy, load_backends_classes


class ExternalAuthMixin(object):
    strategy = None
    backends = {}
    backend = None

    def __init__(self, *args, **kwargs):
        self.init_backends()
        super(ExternalAuthMixin, self).__init__(*args, **kwargs)

    def init_backends(self):
        backend_paths = list(map(lambda x: x.get('backend_path'), settings.SSO_APPLICATIONS.values()))
        self.backends = load_backends_classes(backend_paths)

    def init_auth(self, request, app):
        redirect_uri = self.redirect_uri(app)

        try:
            name = configuration.clean_sso_application_name(app)
            config = settings.SSO_APPLICATIONS[name]
        except KeyError:
            raise NotFound

        backend_path = config.get('backend_path')
        Backend = self.backends.get(backend_path)

        self.strategy = load_strategy(self, request, config)
        self.backend = Backend(self.strategy, redirect_uri)

    def redirect_uri(self, app):
        return '/api/external_auth/complete/{0}/'.format(app)



from jet_bridge_base.configuration import configuration
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.track_database import track_database_async
from jet_bridge_base.utils.track_model import track_model_async


class UpdateAPIViewMixin(object):

    def update(self, request, *args, **kwargs):
        track_database_async(request)

        partial = kwargs.pop('partial', False)
        self.apply_timezone(request)
        instance = self.get_object(request)
        serializer = self.get_serializer(request, instance=instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(request, serializer)

        representation_data = serializer.representation_data
        track_model_async(request, kwargs.get('model'), 'update', kwargs.get('pk'), representation_data)

        return JSONResponse(representation_data)

    def put(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def patch(self, *args, **kwargs):
        self.update(partial=True, *args, **kwargs)

    def perform_update(self, request, serializer):
        configuration.on_model_pre_update(request.path_kwargs['model'], serializer.instance)
        instance = serializer.save()
        configuration.on_model_post_update(request.path_kwargs['model'], instance)

    def partial_update(self, *args, **kwargs):
        kwargs['partial'] = True
        return self.update(*args, **kwargs)



import os
import hashlib

from jet_bridge_base.configuration import configuration


class MediaCache(object):
    max_cache_size = 1024 * 1024 * 50
    files = []
    size = 0
    dir = '_jet_cache'

    def __init__(self):
        self.update_files()

    def get_files(self):
        result = []
        if configuration.media_exists(self.dir):
            directories, files = configuration.media_listdir(self.dir)
            for f in files:
                fp = os.path.join(self.dir, f)
                result.append({
                    'path': fp,
                    'size': configuration.media_size(fp)
                })
            self.sort_files(result)
        return result

    def sort_files(self, files):
        files.sort(key=lambda x: configuration.media_get_modified_time(x['path']))

    def get_files_size(self, files):
        total_size = 0
        for file in files:
            total_size += file['size']
        return total_size

    def update_files(self):
        self.files = self.get_files()
        self.size = self.get_files_size(self.files)

    def add_file(self, path):
        absolute_path = cache.full_path(path)
        size = configuration.media_size(absolute_path)

        self.files.append({
            'path': absolute_path,
            'size': size
        })
        self.size += size
        self.sort_files(self.files)

    def clear_cache_if_needed(self):
        while self.size > self.max_cache_size:
            configuration.media_delete(self.files[0]['path'])
            self.size -= self.files[0]['size']
            self.files.remove(self.files[0])

    def filename(self, path):
        extension = os.path.splitext(path)[1]
        return '{}{}'.format(hashlib.sha256(path.encode('utf8')).hexdigest(), extension)

    def full_path(self, path):
        return os.path.join(self.dir, self.filename(path))

    def exists(self, path):
        thumbnail_full_path = self.full_path(path)
        return configuration.media_exists(thumbnail_full_path)

    def url(self, path, request):
        return configuration.media_url(os.path.join(self.dir, self.filename(path)), request)

cache = MediaCache()



import six

from jet_bridge_base.fields.field import Field


class FloatField(Field):
    field_error_messages = {
        'invalid': 'not a valid float'
    }

    def to_internal_value_item(self, value):
        if value is None:
            return
        value = six.text_type(value).strip()

        if value.lower() == 'true':
            return 1
        elif value.lower() == 'false':
            return 0

        try:
            return float(value)
        except (ValueError, TypeError):
            self.error('invalid')

    def to_representation_item(self, value):
        if value is None:
            return
        return value




class Response(object):
    def __init__(self, data=None, status=None, headers=None, exception=False, content_type=None):
        self.data = data
        self.status = status
        self.exception = exception
        self.content_type = content_type
        self.headers = self.default_headers()

        if headers:
            self.headers.update(headers)

    def default_headers(self):
        return {'Content-Type': 'text/html'}

    def header_items(self):
        if not self.headers:
            return []
        return self.headers.items()

    def render(self):
        return self.data



import json
import jwt
from jwt import PyJWTError

from jet_bridge_base import settings
from jet_bridge_base.utils.compress import decompress_data

USER_TOKEN_PREFIX = 'Token'
PROJECT_TOKEN_PREFIX = 'ProjectToken'
JWT_TOKEN_PREFIX = 'JWT'
BEARER_TOKEN_PREFIX = 'Bearer'


def parse_token(value):
    tokens = value.split(',') if value else []
    result = {}

    for token in tokens:
        try:
            type, data = token.split(' ', 2)
            items = data.split(';')

            if len(items) == 0:
                continue

            try:
                params = dict(map(lambda x: x.split('=', 2), items[1:]))
            except ValueError:
                params = {}

            result[type] = {
                'type': type,
                'value': items[0],
                'params': params
            }
        except (ValueError, AttributeError):
            pass

    if JWT_TOKEN_PREFIX in result:
        return result[JWT_TOKEN_PREFIX]
    elif len(result):
        return list(result.values())[0]


def decode_jwt_token(token, verify_exp=True):
    JWT_VERIFY_KEY = '\n'.join([line.lstrip() for line in settings.JWT_VERIFY_KEY.split('\\n')])

    try:
        return jwt.decode(token, key=JWT_VERIFY_KEY, algorithms=['RS256'], options={'verify_exp': verify_exp})
    except PyJWTError:
        return None


def decompress_permissions(permissions):
    decoded = decompress_data(permissions)
    return json.loads(decoded)



import hashlib
import random
import time

try:
    random = random.SystemRandom()
    using_sysrandom = True
except NotImplementedError:
    import warnings
    warnings.warn('A secure pseudo-random number generator is not available '
                  'on your system. Falling back to Mersenne Twister.')
    using_sysrandom = False


def get_random_string(length, allowed_chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789', salt=''):
    """
    Returns a securely generated random string.

    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    if not using_sysrandom:
        # This is ugly, and a hack, but it makes things better than
        # the alternative of predictability. This re-seeds the PRNG
        # using a value that is hard for an attacker to predict, every
        # time a random string is required. This may change the
        # properties of the chosen random sequence slightly, but this
        # is better than absolute predictability.
        random.seed(
            hashlib.sha256(
                ("%s%s%s" % (
                    random.getstate(),
                    time.time(),
                    salt)).encode('utf-8')
            ).digest())
    return ''.join(random.choice(allowed_chars) for i in range(length))


def find_index(list, predicate):
    for i, value in enumerate(list):
        if predicate(value, i):
            return i
    return None


# TODO: Fix non dict list items
# TODO: List merge is not universal
def merge(destination, source):
    for key, value in source.items():
        if key == 'params':
            destination[key] = value
        elif isinstance(value, dict):
            node = destination.setdefault(key, {})
            merge(node, value)
        elif isinstance(value, list):
            node = destination.setdefault(key, [])

            for item in value:
                index = find_index(node, lambda x, i: x['db_column'] == item['db_column'])
                if index is None:
                    continue
                node[index]
                merge(node[index], item)
        else:
            destination[key] = value

    return destination


def merge_two_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


def get_set_first(value):
    return next(iter(value))


def any_type_sorter(value):
    if value is None:
        return ''
    return str(value)


def unique(arr):
    result = []
    for item in arr:
        if item not in result:
            result.append(item)
    return result


def flatten(arr):
    result = []
    for item in arr:
        if isinstance(item, list):
            result.extend(item)
        else:
            result.append(item)
    return result


def format_size(num, suffix='B'):
    for unit in ('', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi'):
        if abs(num) < 1024.0:
            return f'{num:3.1f}{unit}{suffix}'
        num /= 1024.0
    return f'{num:.1f}Yi{suffix}'


class CollectionDict(dict):
    def __iter__(self):
        for column in self.values():
            yield column



import requests
from requests import RequestException

from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.logger import logger


def api_method_url(method):
    return '{}/{}'.format(settings.API_BASE_URL, method)


def is_project_token_activated(project_token):
    if not project_token:
        return False

    url = api_method_url('project_tokens/{}/'.format(project_token))
    headers = {
        'User-Agent': '{} v{}'.format(configuration.get_type(), configuration.get_version())
    }

    r = requests.request('GET', url, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        return False

    result = r.json()

    return bool(result.get('activated'))


def is_resource_token_activated(project_name, resource_token):
    if not project_name or not resource_token:
        return False

    url = api_method_url('check_resource_token/')
    headers = {
        'User-Agent': '{} v{}'.format(configuration.get_type(), configuration.get_version())
    }
    data = {
        'project': project_name,
        'token': resource_token
    }

    r = requests.request('POST', url, headers=headers, data=data)

    if 200 <= r.status_code < 300:
        result = r.json()
        return bool(result.get('activated'))
    elif 400 <= r.status_code < 500:
        return False
    else:
        raise RequestException()


def project_auth(token, project_token, permission=None, params=None):
    if not project_token:
        return {
            'result': False
        }

    url = api_method_url('project_auth/')
    data = {
        'project_token': project_token,
        'token': token
    }
    headers = {
        'User-Agent': '{} v{}'.format(configuration.get_type(), configuration.get_version())
    }

    if permission:
        data.update(permission)

    if params:
        if 'project_child' in params:
            data['project_child'] = params['project_child']

    r = requests.request('POST', url, data=data, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        logger.error('Project Auth request error: %d %s %s', r.status_code, r.reason, r.text)
        return {
            'result': False
        }

    result = r.json()

    if result.get('access_disabled'):
        return {
            'result': False,
            'warning': result.get('warning')
        }

    return {
        'result': True,
        'warning': result.get('warning')
    }


def get_resource_secret_tokens(project, resource, token):
    if not token:
        return []

    url = api_method_url('projects/{}/resources/{}/secret_tokens/'.format(project, resource))
    headers = {
        'Authorization': 'ProjectToken {}'.format(token),
        'User-Agent': '{} v{}'.format(configuration.get_type(), configuration.get_version())
    }

    r = requests.request('GET', url, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        return []

    return r.json()


def get_secret_tokens(project_name, environment_name, resource, draft, token, user_token):
    if not token:
        return []

    if environment_name:
        method_url = 'projects/{}/{}/secret_tokens/'.format(project_name, environment_name)
    else:
        method_url = 'projects/{}/secret_tokens/'.format(project_name)

    url = api_method_url(method_url)
    headers = {
        'Authorization': 'ProjectToken {}'.format(token),
        'User-Agent': '{} v{}'.format(configuration.get_type(), configuration.get_version())
    }
    data = {
        'resource': resource,
        'user_token': user_token
    }

    if draft:
        data['draft'] = 1

    r = requests.request('POST', url, headers=headers, data=data)
    success = 200 <= r.status_code < 300

    if not success:
        return []

    return r.json()



import time

from jet_bridge_base.configuration import configuration
from jet_bridge_base.db import connections, pending_connections
from jet_bridge_base.db_types import inspect_uniform
from jet_bridge_base.permissions import AdministratorPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.sentry import sentry_controller
from jet_bridge_base.utils.classes import issubclass_safe
from jet_bridge_base.utils.common import format_size
from jet_bridge_base.utils.graphql import ModelFiltersType, ModelFiltersFieldType, ModelFiltersRelationshipType, \
    ModelLookupsType, ModelLookupsFieldType, ModelLookupsRelationshipType, ModelSortType, ModelAttrsType
from jet_bridge_base.utils.process import get_memory_usage
from jet_bridge_base.views.base.api import BaseAPIView


class StatusView(BaseAPIView):
    permission_classes = (AdministratorPermissions,)

    def map_connection_graphql_schema(self, schema):
        if not schema:
            return {'status': 'no_schema'}

        instance = schema.get('instance')
        tables_processed = schema.get('tables_processed', 0)
        tables_total = schema.get('tables_total')

        if instance:
            types_count = len(instance._type_map.values())
            filters_count = 0
            filters_fields_count = 0
            filters_relationships_count = 0
            lookups_count = 0
            lookups_fields_count = 0
            lookups_relationships_count = 0
            sort_count = 0
            attrs_count = 0
            get_schema_time = schema.get('get_schema_time')
            memory_usage_approx = schema.get('memory_usage_approx')

            for item in instance._type_map.values():
                if not hasattr(item, 'graphene_type'):
                    continue

                if issubclass_safe(item.graphene_type, ModelFiltersType):
                    filters_count += 1
                elif issubclass_safe(item.graphene_type, ModelFiltersFieldType):
                    filters_fields_count += 1
                elif issubclass_safe(item.graphene_type, ModelFiltersRelationshipType):
                    filters_relationships_count += 1
                elif issubclass_safe(item.graphene_type, ModelLookupsType):
                    lookups_count += 1
                elif issubclass_safe(item.graphene_type, ModelLookupsFieldType):
                    lookups_fields_count += 1
                elif issubclass_safe(item.graphene_type, ModelLookupsRelationshipType):
                    lookups_relationships_count += 1
                elif issubclass_safe(item.graphene_type, ModelSortType):
                    sort_count += 1
                elif issubclass_safe(item.graphene_type, ModelAttrsType):
                    attrs_count += 1

            return {
                'status': 'ok',
                'tables_processed': tables_processed,
                'tables_total': tables_total,
                'types': types_count,
                'filters': filters_count,
                'filters_fields': filters_fields_count,
                'filters_relationships': filters_relationships_count,
                'lookups': lookups_count,
                'lookups_fields': lookups_fields_count,
                'lookups_relationships': lookups_relationships_count,
                'sort': sort_count,
                'attrs': attrs_count,
                'get_schema_time': get_schema_time,
                'memory_usage_approx': memory_usage_approx,
                'memory_usage_approx_str': format_size(memory_usage_approx) if memory_usage_approx else None
            }
        else:
            return {
                'status': 'pending',
                'tables_processed': tables_processed,
                'tables_total': tables_total
            }

    def map_tunnel(self, tunnel):
        if not tunnel:
            return

        return {
            'is_active': tunnel.is_active,
            'local_address': '{}:{}'.format(tunnel.local_bind_host, tunnel.local_bind_port),
            'remote_address': '{}:{}'.format(tunnel.ssh_host, tunnel.ssh_port)
        }

    def map_connection(self, connection):
        cache = connection['cache']
        MappedBase = connection['MappedBase']
        column_count = 0
        relationships_count = 0

        for Model in MappedBase.classes:
            try:
                mapper = inspect_uniform(Model)
                column_count += len(mapper.columns)
                relationships_count += len(mapper.relationships)
            except Exception as e:
                sentry_controller.capture_exception(e)

        graphql_schema = self.map_connection_graphql_schema(cache.get('graphql_schema'))
        graphql_schema_draft = self.map_connection_graphql_schema(cache.get('graphql_schema_draft'))
        tunnel = self.map_tunnel(connection.get('tunnel'))
        last_request = connection.get('last_request')
        default_timezone_updated = connection.get('default_timezone_updated')

        reflect_memory_usage_approx = connection.get('reflect_memory_usage_approx')

        return {
            'id': connection['id'],
            'name': connection['name'],
            'params_id': connection['params_id'],
            'project': connection.get('project'),
            'token': connection.get('token'),
            'tables': len(MappedBase.classes),
            'columns': column_count,
            'relationships': relationships_count,
            'graphql_schema': graphql_schema,
            'graphql_schema_draft': graphql_schema_draft,
            'init_start': connection.get('init_start'),
            'connect_time': connection.get('connect_time'),
            'reflect_time': connection.get('reflect_time'),
            'reflect_memory_usage_approx': reflect_memory_usage_approx,
            'reflect_memory_usage_approx_str': format_size(reflect_memory_usage_approx) if reflect_memory_usage_approx else None,
            'reflect_metadata_dump': connection.get('reflect_metadata_dump'),
            'default_timezone': str(connection['default_timezone']) if connection.get('default_timezone') else None,
            'default_timezone_updated': default_timezone_updated.isoformat() if default_timezone_updated else None,
            'tunnel': tunnel,
            'last_request': last_request.isoformat() if last_request else None
        }

    def map_pending_connection(self, pending_connection):
        tunnel = self.map_tunnel(pending_connection.get('tunnel'))

        return {
            'id': pending_connection['id'],
            'name': pending_connection['name'],
            'project': pending_connection.get('project'),
            'token': pending_connection.get('token'),
            'init_start': pending_connection.get('init_start'),
            'tables_processed': pending_connection.get('tables_processed', 0),
            'tables_total': pending_connection.get('tables_total'),
            'tunnel': tunnel
        }

    def get(self, request, *args, **kwargs):
        now = time.time()
        uptime = round(now - configuration.init_time)
        memory_used = get_memory_usage()

        active_connections = []
        schema_generating_connections = []

        for connection in connections.values():
            cache = connection['cache']
            graphql_schema = cache.get('graphql_schema')
            graphql_schema_draft = cache.get('graphql_schema_draft')

            if graphql_schema and not graphql_schema.get('instance'):
                schema_generating_connections.append(connection)
            elif graphql_schema_draft and not graphql_schema_draft.get('instance'):
                schema_generating_connections.append(connection)
            else:
                active_connections.append(connection)

        return JSONResponse({
            'total_pending_connections': len(pending_connections.keys()),
            'total_schema_generating_connections': len(schema_generating_connections),
            'total_active_connections': len(active_connections),
            'pending_connections': map(lambda x: self.map_pending_connection(x), pending_connections.values()),
            'schema_generating_connections': map(lambda x: self.map_connection(x), schema_generating_connections),
            'active_connections': map(lambda x: self.map_connection(x), active_connections),
            'memory_used': memory_used,
            'memory_used_str': format_size(memory_used),
            'uptime': uptime
        })



import base64
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from hashlib import sha256

backend = default_backend()


def decrypt(message_encrypted, secret_key):
    message_salt = message_encrypted[-24:]
    message_payload = message_encrypted[:-24]
    salt = base64.b64decode(message_salt)
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=backend
    )
    key = base64.urlsafe_b64encode(kdf.derive(bytes(secret_key, encoding='utf8')))
    f = Fernet(key)
    return f.decrypt(bytes(message_payload, encoding='latin1')).decode('utf8')


def get_sha256_hash(value):
    return sha256(value.encode('utf-8')).hexdigest()



import contextlib
import threading
from datetime import timedelta, datetime

from jet_bridge_base import settings
from jet_bridge_base.db_types import dump_metadata_file, load_mapped_base, init_database_connection, \
    fetch_default_timezone
from jet_bridge_base.logger import logger
from jet_bridge_base.ssh_tunnel import SSHTunnel
from jet_bridge_base.utils.common import get_random_string, format_size
from jet_bridge_base.utils.conf import get_connection_id, get_connection_schema, get_connection_name, \
    get_connection_params_id, is_tunnel_connection, get_conf, get_settings_conf
from jet_bridge_base.utils.datetime import date_trunc_minutes

try:
    from geoalchemy2 import types
except ImportError:
    pass

connections = {}
pending_connections = {}
MODEL_DESCRIPTIONS_RESPONSE_CACHE_KEY = 'model_descriptions_response'
MODEL_DESCRIPTIONS_HASH_CACHE_KEY = 'model_descriptions_hash'


def get_connection_tunnel(conf):
    if not is_tunnel_connection(conf):
        return

    schema = get_connection_schema(conf)
    connection_name = get_connection_name(conf, schema)

    # from sshtunnel import SSHTunnelForwarder, address_to_str
    # import paramiko
    #
    # class SafeSSHTunnelForwarder(SSHTunnelForwarder):
    #     skip_tunnel_checkup = False
    #
    #     def check_is_running(self):
    #         try:
    #             while True:
    #                 time.sleep(5)
    #                 if not tunnel.local_is_up(tunnel.local_bind_address):
    #                     logger.info('SSH tunnel is down, disposing connection "{}"'.format(connection_name))
    #                     break
    #         finally:
    #             dispose_connection(conf)
    #
    #     def start(self):
    #         super(SafeSSHTunnelForwarder, self).start()
    #
    #         for _srv in self._server_list:
    #             thread = threading.Thread(
    #                 target=self.check_is_running,
    #                 args=(),
    #                 name='Srv-{0}-check'.format(address_to_str(_srv.local_port))
    #             )
    #             thread.start()
    #
    # private_key_str = conf.get('ssh_private_key').replace('\\n', '\n')
    # private_key = paramiko.RSAKey.from_private_key(StringIO(private_key_str))
    #
    # tunnel = SafeSSHTunnelForwarder(
    #     ssh_address_or_host=(conf.get('ssh_host'), int(conf.get('ssh_port'))),
    #     ssh_username=conf.get('ssh_user'),
    #     ssh_pkey=private_key,
    #     remote_bind_address=(conf.get('host'), int(conf.get('port'))),
    #     logger=logger
    # )
    # tunnel.start()
    #
    # return tunnel

    logger.info('Starting SSH tunnel for connection "{}"...'.format(connection_name))

    def on_close():
        connection_id = get_connection_id(conf)
        connection = connections.get(connection_id)

        if connection:
            logger.info('SSH tunnel is closed, disposing connection "{}"'.format(connection_name))
            dispose_connection(conf)
        else:
            logger.info('SSH tunnel is closed for connection "{}"'.format(connection_name))

    tunnel = SSHTunnel(
        name=connection_name,
        ssh_host=conf.get('ssh_host'),
        ssh_port=conf.get('ssh_port'),
        ssh_user=conf.get('ssh_user'),
        ssh_private_key=conf.get('ssh_private_key').replace('\\n', '\n'),
        remote_host=conf.get('host'),
        remote_port=conf.get('port'),
        on_close=on_close
    )
    tunnel.start()

    logger.info('SSH tunnel started on port {} for connection "{}"'.format(tunnel.local_bind_port, connection_name))

    return tunnel


def wait_pending_connection(connection_id, connection_name):
    pending_connection = pending_connections.get(connection_id)
    if not pending_connection:
        return

    logger.info('Waiting database connection "{}"...'.format(connection_name))

    connected_condition = pending_connection['connected']
    with connected_condition:
        timeout = timedelta(minutes=10).total_seconds()
        connected_condition.wait(timeout=timeout)

    connection = connections.get(connection_id)
    if connection:
        logger.info('Found database connection "{}"'.format(connection_name))
        return connection
    else:
        logger.info('Not found database connection "{}"'.format(connection_name))


def clean_hostname(hostname):
    if not isinstance(hostname, str):
        return
    hostname = hostname.strip().lower()
    if hostname == '':
        return
    return hostname


def get_blacklist_hostnames():
    hostnames = []

    if settings.BLACKLIST_HOSTS:
        hostnames.extend(settings.BLACKLIST_HOSTS.split(','))

    try:
        import configparser
    except ImportError:
        import ConfigParser as configparser

    try:
        config = configparser.RawConfigParser()
        config.read(settings.CONFIG)

        config_value = config.get('JET', 'BLACKLIST_HOSTS', fallback='')
        hostnames.extend(config_value.split(','))
    except:
        pass

    return list(filter(
        lambda x: x is not None,
        map(lambda x: clean_hostname(x), hostnames)
    ))


def is_hostname_blacklisted(hostname):
    hostname = clean_hostname(hostname)
    if not hostname:
        return False

    blacklist_hosts = get_blacklist_hostnames()
    if len(blacklist_hosts) == 0:
        return False

    return hostname in blacklist_hosts


def connect_database(conf):
    global connections, pending_connections

    hostname = conf.get('host')
    if is_hostname_blacklisted(hostname):
        raise Exception('Hostname "{}" is blacklisted'.format(hostname))

    connection_id = get_connection_id(conf)
    connection_params_id = get_connection_params_id(conf)
    schema = get_connection_schema(conf)
    connection_name = get_connection_name(conf, schema)
    id_short = connection_id[:4]

    existing_connection = connections.get(connection_id)
    if existing_connection:
        if existing_connection['params_id'] == connection_params_id:
            return existing_connection
        else:
            logger.info('[{}] Reconnecting to database "{}" because of different params ({} {})...'.format(
                id_short,
                connection_name,
                connection_params_id,
                existing_connection['params_id']
            ))
            dispose_connection(conf)

    init_start = datetime.now()

    connected_condition = threading.Condition()
    pending_connection_id = get_random_string(32)
    pending_connection = {
        'id': pending_connection_id,
        'name': connection_name,
        'project': conf.get('project'),
        'token': conf.get('token'),
        'init_start': init_start.isoformat(),
        'connected': connected_condition
    }

    existing_connection = wait_pending_connection(connection_id, connection_name)
    if existing_connection:
        return existing_connection

    pending_connections[connection_id] = pending_connection
    tunnel = None

    try:
        tunnel = get_connection_tunnel(conf)
        pending_connection['tunnel'] = tunnel

        database_connection = init_database_connection(
            conf,
            tunnel,
            id_short,
            connection_name,
            schema,
            pending_connection
        )

        connections[connection_id] = {
            'id': connection_id,
            'name': connection_name,
            'params_id': connection_params_id,
            'tunnel': tunnel,
            'cache': {},
            'lock': threading.Lock(),
            'project': conf.get('project'),
            'token': conf.get('token'),
            'init_start': init_start.isoformat(),
            'last_request': datetime.now(),
            **database_connection
        }

        return connections[connection_id]
    except Exception as e:
        if tunnel:
            tunnel.close()

        raise e
    finally:
        if connection_id in pending_connections and pending_connections[connection_id].get('id') == pending_connection_id:
            del pending_connections[connection_id]

        with connected_condition:
            connected_condition.notify_all()


def dispose_connection_object(connection):
    try:
        connection['engine'].dispose()

        if connection.get('tunnel'):
            connection['tunnel'].close()

        return True
    except Exception:
        return False


def dispose_connection(conf):
    global connections

    connection_id = get_connection_id(conf)
    connection = connections.get(connection_id)

    if connection and dispose_connection_object(connection):
        del connections[connection_id]
        return True

    return False


def dispose_request_connection(request):
    return dispose_connection(get_conf(request))


def get_connection(request):
    conf = get_conf(request)
    connection_id = get_connection_id(conf)
    return connections.get(connection_id)


def connect_database_from_settings():
    if settings.DATABASE_ENGINE == 'none':
        return
    return connect_database(get_settings_conf())


def get_request_connection(request):
    return connect_database(get_conf(request))


def create_session(request):
    connection = get_request_connection(request)
    if not connection:
        return

    conf = get_conf(request)
    hostname = conf.get('host')

    if is_hostname_blacklisted(hostname):
        dispose_connection(conf)
        raise Exception('Hostname "{}" is blacklisted'.format(hostname))

    return connection['Session']()


def get_connection_id_short(request):
    connection = get_request_connection(request)
    if not connection or 'id' not in connection:
        return
    return connection['id'][:4]


def get_mapped_base(request):
    connection = get_request_connection(request)
    if not connection:
        return
    return connection['MappedBase']


def get_engine(request):
    connection = get_request_connection(request)
    if not connection:
        return
    return connection['engine']


def get_type_code_to_sql_type(request):
    connection = get_request_connection(request)
    if not connection:
        return
    return connection['type_code_to_sql_type']


def get_default_timezone(request):
    connection = get_request_connection(request)

    if not connection:
        return

    default_timezone = connection.get('default_timezone')
    default_timezone_updated = connection.get('default_timezone_updated')

    if not default_timezone:
        return

    hour_now = date_trunc_minutes(datetime.now())
    hour_timezone_updated = date_trunc_minutes(default_timezone_updated)

    if hour_now.timestamp() != hour_timezone_updated.timestamp():
        conf = get_conf(request)
        connection_id = get_connection_id(conf)
        id_short = connection_id[:4]

        new_default_timezone = fetch_default_timezone(request.session)
        new_default_timezone_updated = datetime.now()

        if new_default_timezone is not None:
            connection['default_timezone'] = new_default_timezone
            connection['default_timezone_updated'] = new_default_timezone_updated

            logger.info('[{}] Default timezone updated: "{}"'.format(id_short, default_timezone))

            return new_default_timezone
        else:
            logger.info('[{}] Failed to update default timezone'.format(id_short))

    return default_timezone


@contextlib.contextmanager
def connection_cache(connection):
    if not connection:
        yield {}
    with connection['lock']:
        yield connection['cache']


@contextlib.contextmanager
def request_connection_cache(request):
    connection = get_request_connection(request)
    if not connection:
        yield {}
    with connection['lock']:
        yield connection['cache']


def connection_cache_get(request, name, default=None):
    connection = get_request_connection(request)
    if not connection:
        return
    with connection['lock']:
        return connection['cache'].get(name, default)


def connection_cache_set(request, name, value):
    connection = get_request_connection(request)
    if not connection:
        return
    with connection['lock']:
        connection['cache'][name] = value


def reload_request_mapped_base(request):
    conf = get_conf(request)
    MappedBase = get_mapped_base(request)

    load_mapped_base(MappedBase, True)
    reload_request_model_descriptions_cache(request)
    reload_request_graphql_schema(request)
    dump_metadata_file(conf, MappedBase.metadata)


def reload_connection_graphql_schema(connection, draft=None):
    with connection_cache(connection) as cache:
        if draft is None:
            cache['graphql_schema'] = None
            cache['graphql_schema_draft'] = None
        else:
            schema_key = 'graphql_schema_draft' if draft else 'graphql_schema'
            cache[schema_key] = None


def reload_request_graphql_schema(request, draft=None):
    with request_connection_cache(request) as cache:
        if draft is None:
            cache['graphql_schema'] = None
            cache['graphql_schema_draft'] = None
        else:
            schema_key = 'graphql_schema_draft' if draft else 'graphql_schema'
            cache[schema_key] = None


def reload_request_model_descriptions_cache(request):
    with request_connection_cache(request) as cache:
        cache[MODEL_DESCRIPTIONS_RESPONSE_CACHE_KEY] = None
        cache[MODEL_DESCRIPTIONS_HASH_CACHE_KEY] = None


def release_inactive_graphql_schemas():
    if not settings.RELEASE_INACTIVE_GRAPHQL_SCHEMAS_TIMEOUT:
        return

    for connection in connections.values():
        cache = connection['cache']
        graphql_schema = cache.get('graphql_schema')
        graphql_schema_draft = cache.get('graphql_schema_draft')

        if not graphql_schema and not graphql_schema_draft:
            continue

        time_elapsed = (datetime.now() - connection['last_request']).total_seconds()

        if time_elapsed <= settings.RELEASE_INACTIVE_GRAPHQL_SCHEMAS_TIMEOUT:
            continue

        graphql_schema_memory = graphql_schema.get('memory_usage_approx') if graphql_schema else 0
        graphql_schema_draft_memory = graphql_schema_draft.get('memory_usage_approx') if graphql_schema_draft else 0
        memory_usage_approx = graphql_schema_memory + graphql_schema_draft_memory

        logger.info('Release inactive GraphQL schema "{}" (MEM:{}, ELAPSED:{})...'.format(
            connection['name'],
            format_size(memory_usage_approx) if memory_usage_approx else None,
            '{}s'.format(round(time_elapsed))
        ))

        reload_connection_graphql_schema(connection)



from __future__ import absolute_import
import json

from six import string_types

from jet_bridge_base.fields.field import Field


class JSONField(Field):
    field_error_messages = {
        'invalid': 'not a valid JSON'
    }

    def __init__(self, *args, **kwargs):
        if 'allow_many' not in kwargs:
            kwargs['allow_many'] = True
        super(JSONField, self).__init__(*args, **kwargs)

    def to_internal_value_item(self, value):
        if isinstance(value, string_types):
            try:
                return json.loads(value)
            except ValueError:
                self.error('invalid')
        else:
            return value

    def to_representation_item(self, value):
        return value



import json
import os

from jet_bridge_base import settings
from jet_bridge_base.logger import logger
from jet_bridge_base.utils.conf import get_connection_id, get_connection_schema, get_connection_name, get_metadata_file_path

from .mongo_metadata import MongoMetadata


def mongo_dump_metadata_file(conf, metadata):
    if not settings.CACHE_METADATA:
        return

    connection_id = get_connection_id(conf)
    schema = get_connection_schema(conf)
    connection_name = get_connection_name(conf, schema)
    id_short = connection_id[:4]

    file_path = get_metadata_file_path(conf)

    try:
        dir_path = os.path.dirname(file_path)

        if dir_path and not os.path.exists(dir_path):
            os.makedirs(dir_path)

        with open(file_path, 'w') as file:
            json.dump(metadata.serialize(), file, cls=MetadataJSONEncoder)

        logger.info('[{}] Saved schema cache for "{}"'.format(id_short, connection_name))

        return file_path
    except Exception as e:
        logger.error('[{}] Failed dumping schema cache for "{}"'.format(id_short, connection_name), exc_info=e)


class MetadataJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return super(MetadataJSONEncoder, self).default(obj)


def mongo_load_metadata_file(conf):
    if not settings.CACHE_METADATA:
        return

    connection_id = get_connection_id(conf)
    schema = get_connection_schema(conf)
    connection_name = get_connection_name(conf, schema)
    id_short = connection_id[:4]

    file_path = get_metadata_file_path(conf)

    if not os.path.exists(file_path):
        logger.info('[{}] Schema cache not found for "{}"'.format(id_short, connection_name))
        return

    try:
        with open(file_path, 'r') as file:
            data = json.load(file)
            metadata = MongoMetadata.deserialize(data)

        return {
            'file_path': file_path,
            'metadata': metadata
        }
    except Exception as e:
        logger.error('[{}] Failed loading schema cache for "{}"'.format(id_short, connection_name), exc_info=e)



from social_core.utils import build_absolute_uri
from social_core.strategy import BaseStrategy

from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.responses.redirect import RedirectResponse
from jet_bridge_base.utils.common import merge_two_dicts


class JetBridgeStrategy(BaseStrategy):

    def __init__(self, storage, request_handler, request, config, tpl=None):
        self.request_handler = request_handler
        self.request = request
        self.config = config
        super(JetBridgeStrategy, self).__init__(storage, tpl)

    def authenticate(self, *args, **kwargs):
        result = super(JetBridgeStrategy, self).authenticate(*args, **kwargs)
        # Ignore redirect response by returning user of another class
        return {
            'auth': True,
            'details': result.details,
            'extra_data': result.extra_data
        } if result else {}

    def get_setting(self, name):
        settings = merge_two_dicts(self.config, {
            'pipeline': [
                'social_core.pipeline.social_auth.social_details',
                'jet_bridge_base.external_auth.pipeline.save_extra_data',
                'jet_bridge_base.external_auth.pipeline.return_result'
            ]
        })
        key = name.lower()
        return settings[key]

    def request_data(self, merge=True):
        return merge_two_dicts(self.request.data, self.request.query_arguments)

    def request_host(self):
        return self.request.host

    def redirect(self, url):
        return RedirectResponse(url)

    def html(self, content):
        self.request_handler.write(content)

    def session_get(self, name, default=None, secure=True):
        value = configuration.session_get(self.request, name, default, secure=secure)
        if value:
            return value
        return default

    def session_set(self, name, value, secure=True):
        configuration.session_set(self.request, name, value, secure=secure)

    def session_pop(self, name):
        value = self.session_get(name)
        configuration.session_clear(self.request, name)
        return value

    def session_setdefault(self, name, value):
        pass

    def build_absolute_uri(self, path=None):
        if settings.BASE_URL is None:
            base_url = '{0}://{1}'.format(self.request.protocol, self.request.host)
        else:
            base_url = settings.BASE_URL

        return build_absolute_uri(
            base_url,
            path
        )



import six

from jet_bridge_base.fields.field import Field


class WKTField(Field):
    field_error_messages = {
        'invalid': 'not a valid Geo object - {error}'
    }

    def to_internal_value_item(self, value):
        if value is None:
            return
        from geoalchemy2 import ArgumentError, WKTElement
        try:
            return WKTElement(value)
        except ArgumentError as e:
            self.error('invalid', error=six.text_type(e))

    def to_representation_item(self, value):
        if value is None:
            return
        from geoalchemy2.shape import to_shape
        shape = to_shape(value)
        if hasattr(shape, 'wkt'):
            return shape.wkt
        else:
            # Backward compatibility Shapely <1.8.0
            return shape.to_wkt()



import platform
from datetime import datetime
import sys

import six

from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.db import create_session
from jet_bridge_base.exceptions.api import APIException
from jet_bridge_base.exceptions.not_found import NotFound
from jet_bridge_base.exceptions.permission_denied import PermissionDenied
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.responses.template import TemplateResponse
from jet_bridge_base.logger import logger
from jet_bridge_base.utils.common import format_size
from jet_bridge_base.utils.conf import get_connection_name, get_conf, get_connection_schema, get_connection_id
from jet_bridge_base.utils.exceptions import serialize_validation_error


class BaseAPIView(object):
    # request = None
    # session = None
    permission_classes = []
    track_queries = False

    def log_request(self, request):
        params = {'IP': request.get_ip(), 'SID': request.get_stick_session()}
        params_str = ' '.join(map(lambda x: '='.join([x[0], x[1]]), filter(lambda x: x[1], params.items())))
        logger.debug('{} {} {}'.format(request.method, request.full_url(), params_str))

    def before_dispatch(self, request):
        self.track_query_start(request)
        self.log_request(request)

        method_override = request.headers.get('X_HTTP_METHOD_OVERRIDE')
        if method_override is not None:
            request.method = method_override

        self.before_dispatch_permissions_check(request)

    def before_dispatch_permissions_check(self, request):
        if request.method != 'OPTIONS':
            self.check_permissions(request)

    def after_dispatch(self, request):
        self.track_query_finish(request)

    def is_track_queries_enabled(self):
        return self.track_queries and (settings.TRACK_QUERY_SLOW_TIME or settings.TRACK_QUERY_HIGH_MEMORY)

    def track_query_start(self, request):
        if not self.is_track_queries_enabled():
            return

        request.start_track()

    def track_query_finish(self, request):
        if not self.is_track_queries_enabled():
            return

        track_time = request.get_track_time()
        track_memory_usage = request.get_track_memory_usage()

        tags = []

        if track_time >= settings.TRACK_QUERY_SLOW_TIME:
            tags.append('SLOW')

        if track_memory_usage >= settings.TRACK_QUERY_HIGH_MEMORY:
            tags.append('HIGHMEM')

        if not len(tags):
            return

        conf = get_conf(request)
        schema = get_connection_schema(conf)
        connection_id = get_connection_id(conf)
        connection_name = get_connection_name(conf, schema)
        id_short = connection_id[:4]

        track_query_name = self.get_track_query_name(request) or '{} {}'.format(request.method, request.path)

        logger.info('[{}] {} [{}] (MEM:{}, TIME:{}) "{}"'.format(
            id_short,
            track_query_name,
            ','.join(tags),
            format_size(track_memory_usage) if track_memory_usage else None,
            '{}s'.format(track_time),
            connection_name
        ))

    def get_track_query_name(self, request):
        pass

    def on_finish(self):
        pass

    def get_permissions(self):
        return [permission() for permission in self.permission_classes]

    def check_permissions(self, request):
        if settings.DISABLE_AUTH:
            return

        for permission in self.get_permissions():
            if not permission.has_permission(self, request):
                raise PermissionDenied(getattr(permission, 'message', 'forbidden'))

    def check_object_permissions(self, request, obj):
        if settings.DISABLE_AUTH:
            return

        for permission in self.get_permissions():
            if not permission.has_object_permission(self, request, obj):
                raise PermissionDenied(getattr(permission, 'message', 'forbidden'))

    def default_headers(self):
        headers = {}

        if settings.CORS_HEADERS:
            headers['Access-Control-Allow-Origin'] = settings.ALLOW_ORIGIN
            headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, PATCH, DELETE, OPTIONS'
            headers['Access-Control-Allow-Headers'] = 'Authorization,DNT,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type,Range,X-Application-Warning,X-HTTP-Method-Override,X-Bridge-Settings,X-Stick-Session'
            headers['Access-Control-Expose-Headers'] = 'Content-Length,Content-Range,Content-Disposition,Content-Type,X-Application-Warning'
            headers['Access-Control-Allow-Credentials'] = 'true'

        return headers

    def error_response(self, request, exc_type, exc, traceback):
        if isinstance(exc, PermissionDenied):
            return TemplateResponse('403.html', status=403, data={
                'path': request.path,
            })
        elif isinstance(exc, NotFound):
            return TemplateResponse('404.html', status=404, data={
                'path': request.path,
            })
        elif isinstance(exc, ValidationError):
            response = serialize_validation_error(exc)
            return JSONResponse(response, status=exc.status_code)
        elif isinstance(exc, APIException):
            return JSONResponse({
                'error': exc.detail,
                'error_code': exc.code
            }, status=exc.status_code)
        else:
            if settings.DEBUG:
                ctx = {
                    'path': request.path if request else None,
                    'full_path': request.protocol + '://' + request.host + request.path if request else None,
                    'method': request.method if request else None,
                    'type': configuration.get_type(),
                    'version': configuration.get_version(),
                    'current_datetime': datetime.now().strftime('%c'),
                    'python_version': platform.python_version(),
                    'python_executable': sys.executable,
                    'python_path': sys.path
                }

                if exc:
                    ctx.update({
                        'exception_type': exc_type.__name__,
                        'exception_value': six.text_type(exc)
                    })

                if traceback:
                    last_traceback = traceback

                    while last_traceback.tb_next:
                        last_traceback = last_traceback.tb_next

                    frame = last_traceback.tb_frame
                    func_name = frame.f_code.co_name
                    file_name = frame.f_code.co_filename
                    line_number = frame.f_lineno

                    ctx.update({
                        'exception_last_traceback_line': line_number,
                        'exception_last_traceback_func': func_name,
                        'exception_last_traceback_file': file_name,
                    })

                logger.exception(exc)

                return TemplateResponse('500.debug.html', status=500, data=ctx)
            else:
                logger.exception(exc)

                return TemplateResponse('500.html', status=500)

    def dispatch(self, action, request, *args, **kwargs):
        if not hasattr(self, action):
            raise NotFound()
        return getattr(self, action)(request, *args, **kwargs)

    # def build_absolute_uri(self, request, url):
    #     return request.protocol + '://' + request.host + url


class APIView(BaseAPIView):

    def before_dispatch(self, request):
        super(APIView, self).before_dispatch(request)

        try:
            request.session = create_session(request)
        except Exception as e:
            raise ValidationError(str(e))

    def after_dispatch(self, request):
        super(APIView, self).after_dispatch(request)

        if request.session is not None:
            request.session.close()
            request.session = None



import json
import time

from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.external_auth.storage import User
from jet_bridge_base.utils.compress import compress_data


def save_extra_data(backend, user, response, details, strategy, *args, **kwargs):
    extra_data = backend.extra_data(user, 0, response, details, *args, **kwargs)
    sso = configuration.clean_sso_application_name(getattr(backend, 'sso'))
    extra_data_key = '_'.join(['extra_data', sso])

    data = {
        'expires_on': extra_data.get('expires_on'),
        'access_token': extra_data.get('access_token'),
        'expires': extra_data.get('expires'),
        'auth_time': extra_data.get('auth_time'),
        'refresh_token': extra_data.get('refresh_token'),
        'token_updated': int(time.time())
    }

    extra_data_str = json.dumps(data)

    if settings.COOKIE_COMPRESS:
        extra_data_str = compress_data(extra_data_str)

    strategy.session_set(extra_data_key, extra_data_str, secure=not settings.COOKIE_COMPRESS)

    return {
        'extra_data': extra_data
    }


def return_result(details, extra_data, *args, **kwargs):
    user = User()
    user.details = details
    user.extra_data = extra_data

    return {
        'user': user
    }



import re
import six

from jet_bridge_base.db_types import inspect_uniform
from jet_bridge_base.exceptions.validation_error import ValidationError


def serialize_validation_error(exc):
    def process(e, root=False):
        if isinstance(e.detail, dict):
            return dict(map(lambda x: (x[0], process(x[1])), e.detail.items()))
        elif isinstance(e.detail, list):
            return list(map(lambda x: process(x), e.detail))
        elif root:
            return {'non_field_errors': [e.detail]}
        else:
            return e.detail

    return process(exc, root=True)


def validation_error_from_database_error(e, model):
    if hasattr(e, 'orig'):
        if hasattr(e.orig, 'args') and hasattr(e.orig.args, '__getitem__'):
            if len(e.orig.args) == 1:
                message = e.orig.args[0]
            elif len(e.orig.args) == 2:
                message = e.orig.args[1]
            else:
                message = e.orig.args

            message = six.text_type(message)

            regex = [
                [r'Key\s\((.+)\)=\((.+)\)\salready\sexists', 1, 2],  # PostgreSQL
                [r'Duplicate\sentry\s\'(.+)\'\sfor key\s\'(.+)\'', 2, 1],  # MySQL
                [r'UNIQUE\sconstraint\sfailed\:\s(.+)\.(.+)', 2, None]  # SQLite
            ]

            for (r, field_index, value_index) in regex:
                match = re.search(r, message, re.IGNORECASE | re.MULTILINE)

                if match:
                    mapper = inspect_uniform(model)
                    columns = dict(map(lambda x: (x.key, x), mapper.columns))
                    column_name = match.group(field_index)

                    if column_name in columns:
                        error = dict()
                        error[column_name] = ValidationError('record with the same value already exists')
                        return ValidationError(error)

            return ValidationError(message)
    return ValidationError('Query failed')



import pymongo
import sqlalchemy
from sqlalchemy import desc, sql, func, or_, cast
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import operators, sqltypes, text
from sqlalchemy.sql.elements import AnnotatedColumnElement, UnaryExpression
from bson import ObjectId

from jet_bridge_base.db_types.mongo import MongoOperator
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.models import data_types
from jet_bridge_base.utils.classes import is_instance_or_subclass

from .common import inspect_uniform, get_session_engine
from .mongo import MongoDeclarativeMeta, MongoColumn, MongoDesc, MongoQueryset


def desc_uniform(column):
    if isinstance(column, MongoColumn):
        return MongoDesc(column)
    else:
        return desc(column)


def empty_filter(cls):
    if isinstance(cls, DeclarativeMeta):
        return sql.false()
    elif isinstance(cls, MongoDeclarativeMeta):
        return cls._id.exists(False)


def get_queryset_order_by(queryset):
    if isinstance(queryset, MongoQueryset):
        return queryset.get_order_by()
    else:
        if hasattr(queryset, '_order_by_clauses') and queryset._order_by_clauses:
            return queryset._order_by_clauses
        elif hasattr(queryset, '_order_by') and queryset._order_by:
            return queryset._order_by
        else:
            return []


def get_queryset_limit(queryset):
    if isinstance(queryset, MongoQueryset):
        return queryset.get_limit()
    else:
        if hasattr(queryset, '_limit_clause') and queryset._limit_clause:
            return queryset._limit_clause
        elif hasattr(queryset, '_limit') and queryset._limit:
            return queryset._limit


def apply_default_ordering(Model, queryset):
    mapper = inspect_uniform(Model)
    pk = mapper.primary_key[0]
    ordering = get_queryset_order_by(queryset)

    def is_pk(x):
        if isinstance(queryset, MongoQueryset):
            return x[0] == pk.name
        else:
            if isinstance(x, AnnotatedColumnElement):
                return x.name == pk.name
            elif isinstance(x, UnaryExpression):
                return hasattr(x.element, 'name') and x.element.name == pk.name and x.modifier == operators.desc_op
        return False

    if ordering is None or not any(map(is_pk, ordering)):
        queryset = queryset.order_by(desc_uniform(pk))

    return queryset


def queryset_count_optimized_for_postgresql(session, db_table):
    try:
        cursor = session.execute(text('SELECT reltuples FROM pg_class WHERE relname = :db_table'), {'db_table': db_table})
        row = cursor.fetchone()
        return int(row[0])
    except SQLAlchemyError:
        session.rollback()
        raise


def queryset_count_optimized_for_mysql(session, db_table):
    try:
        cursor = session.execute(text('EXPLAIN SELECT COUNT(*) FROM `{}`'.format(db_table)))
        row = cursor.fetchone()
        return int(row[8])
    except SQLAlchemyError:
        session.rollback()
        raise


def queryset_count_optimized(session, queryset):
    if queryset.whereclause is None:
        count_optimized = None

        if isinstance(queryset, MongoQueryset):
            count_optimized = queryset.estimated_document_count()
        else:
            try:
                table = queryset.statement.froms[0].name

                if get_session_engine(queryset.session) == 'postgresql':
                    count_optimized = queryset_count_optimized_for_postgresql(session, table)
                elif get_session_engine(queryset.session) == 'mysql':
                    count_optimized = queryset_count_optimized_for_mysql(session, table)
            except:
                pass

        if count_optimized is not None and count_optimized >= 10000:
            return count_optimized

    try:
        return queryset.count()
    except SQLAlchemyError:
        queryset.session.rollback()
        raise


def get_sql_aggregate_func_by_name(name, column):
    if name == 'count':
        return func.count(column)
    elif name == 'sum':
        return func.sum(column)
    elif name == 'min':
        return func.min(column)
    elif name == 'max':
        return func.max(column)
    elif name == 'avg':
        return func.avg(column)


def get_mongo_aggregate_func_by_name(name, column):
    if name == 'count':
        return {'$count': {}}
    elif name == 'sum':
        return {'$sum': '${}'.format(column.name)}
    elif name == 'min':
        return {'$min': '${}'.format(column.name)}
    elif name == 'max':
        return {'$max': '${}'.format(column.name)}
    elif name == 'avg':
        return {'$avg': '${}'.format(column.name)}


def queryset_aggregate(Model, qs, value):
    if isinstance(qs, MongoQueryset):
        y_column = getattr(Model, value['y_column'])
        y_func = get_mongo_aggregate_func_by_name(value['y_func'], y_column)

        if y_func is None:
            return qs.filter(empty_filter(Model))

        result = qs.aggregate({
            '$group': {
                '_id': None,
                'aggregation': y_func
            }
        })

        return result['aggregation']
    else:
        y_column = getattr(Model, value['y_column'])
        y_func = get_sql_aggregate_func_by_name(value['y_func'], y_column)

        if y_func is None:
            return qs.filter(empty_filter(Model))

        whereclause = qs.whereclause
        qs = qs.session.query(y_func)

        if whereclause is not None:
            qs = qs.filter(whereclause)

        return qs.first()[0]

date_trunc_options = {
    'microsecond': 'microseconds',
    'millisecond': 'milliseconds',
    'second': 'minute',
    'minute': 'minute',
    'hour': 'hour',
    'day': 'day',
    'week': 'week',
    'month': 'month',
    'quarter': 'quarter',
    'year': 'year'
}

strftime_options = {
    'microseconds': '%Y-%m-%d %H:%i:%s.%f',
    # 'milliseconds': '%Y-%m-%d %H:%i:%s.%f',
    'second': '%Y-%m-%d %H:%i:%s',
    'minute': '%Y-%m-%d %H:%i:00',
    'hour': '%Y-%m-%d %H:00:00',
    'day': '%Y-%m-%d',
    # 'week': '%Y-%m-%d',
    'month': '%Y-%m-01',
    # 'quarter': '%Y-%m-%d',
    'year': '%Y-01-01'
}

isoformat_options = {
    'microseconds': '%Y-%m-%dT%H:%M:%S.%LZ',
    # 'milliseconds': '%Y-%m-%d %H:%i:%s.%f',
    'second': '%Y-%m-%dT%H:%M:%S.000Z',
    'minute': '%Y-%m-%dT%H:%M:00.000Z',
    'hour': '%Y-%m-%dT%H:00:00.000Z',
    'day': '%Y-%m-%dT00:00:00.000Z',
    # 'week': '%Y-%m-%d',
    'month': '%Y-%m-01T00:00:00.000Z',
    # 'quarter': '%Y-%m-%d',
    'year': '%Y-01-01T00:00:00.000Z',
}

dateadd_options = {
    'millisecond': 'millisecond',
    'second': 'minute',
    'minute': 'minute',
    'hour': 'hour',
    'day': 'day',
    'week': 'week',
    'month': 'month',
    'quarter': 'quarter',
    'year': 'year'
}


def date_trunc_column_clickhouse(column, date_group):
    if date_group == 'microsecond':
        return func.toStartOfMicrosecond(column)
    elif date_group == 'millisecond':
        return func.toStartOfMillisecond(column)
    elif date_group == 'second':
        return func.toStartOfSecond(column)
    elif date_group == 'minute':
        return func.toStartOfMinute(column)
    elif date_group == 'hour':
        return func.toStartOfHour(column)
    elif date_group == 'day':
        return func.toStartOfDay(column)
    elif date_group == 'week':
        return func.toStartOfWeek(column)
    elif date_group == 'month':
        return func.toStartOfMonth(column)
    elif date_group == 'quarter':
        return func.toStartOfQuarter(column)
    elif date_group == 'year':
        return func.toStartOfYear(column)


def get_sql_group_func_lookup(session, lookup_type, lookup_param, column):
    if lookup_type == 'auto':
        field_type = column.property.columns[0].type if hasattr(column, 'property') else column.type

        if is_instance_or_subclass(field_type, (sqltypes.DateTime, sqltypes.Date)):
            lookup_type = 'date'
            lookup_param = 'day'

    try:
        if lookup_type == 'date':
            date_group = lookup_param or 'day'

            if get_session_engine(session) == 'postgresql':
                if date_group in date_trunc_options:
                    return func.date_trunc(date_trunc_options[date_group], column)
            elif get_session_engine(session) == 'mysql':
                if date_group in strftime_options:
                    return func.date_format(column, strftime_options[date_group])
            elif get_session_engine(session) == 'mssql':
                if date_group in dateadd_options:
                    interval = dateadd_options[date_group]
                    return func.dateadd(text(interval), func.datediff(text(interval), text('0'), column), text('0'))
            elif get_session_engine(session) == 'clickhouse':
                date_trunc = date_trunc_column_clickhouse(column, date_group)
                if date_trunc is not None:
                    return date_trunc
            else:
                if date_group in strftime_options:
                    return func.strftime(strftime_options[date_group], column)
        elif lookup_type == 'plain' or lookup_type == 'auto':
            return column
    except IndexError:
        pass

    if lookup_type:
        print('Unsupported lookup: {}'.format(lookup_type))

    return column


def get_mongo_group_func_lookup(lookup_type, lookup_param, column):
    if lookup_type == 'auto':
        if any(map(lambda x: column.type == x, (data_types.DATE, data_types.DATE_TIME))):
            lookup_type = 'date'
            lookup_param = 'day'

    try:
        if lookup_type == 'date':
            date_group = lookup_param or 'day'

            if date_group in isoformat_options:
                return {'$dateToString': {'format': isoformat_options[date_group], 'date': '${}'.format(column.name)}}
        elif lookup_type == 'plain' or lookup_type == 'auto':
            return '${}'.format(column.name)
    except IndexError:
        pass

    if lookup_type:
        print('Unsupported lookup: {}'.format(lookup_type))

    return '${}'.format(column.name)


def queryset_group(Model, qs, value):
    if isinstance(qs, MongoQueryset):
        x_columns = list(map(lambda x: getattr(Model, x), value['x_columns']))
        y_column = getattr(Model, value['y_column'])
        y_func = get_mongo_aggregate_func_by_name(value['y_func'], y_column)

        if y_func is None:
            return qs.filter(empty_filter(Model))

        def group_name(i):
            if i == 0:
                return 'group'
            else:
                return 'group_{}'.format(i + 1)

        def map_lookup(column, i):
            lookup_name = value['x_lookups'][i] if i < len(value['x_lookups']) else None
            lookup_params = lookup_name.split('_') if lookup_name else []
            lookup_type = lookup_params[0] if len(lookup_params) >= 1 else None
            lookup_param = lookup_params[1] if len(lookup_params) >= 2 else None

            if lookup_type == 'date':
                mapper = inspect_uniform(Model)
                column = mapper.columns.get(column.name)
                if column is not None and not any(
                        map(lambda x: column.type == x, (data_types.DATE, data_types.DATE_TIME))):
                    raise ValidationError('Can\'t apply date functions to non-date field: {}'.format(column.name))

            lookup = get_mongo_group_func_lookup(lookup_type, lookup_param, column)

            return {
                'name': group_name(i),
                'lookup': lookup
            }

        x_lookups = list(map(lambda x: map_lookup(x[1], x[0]), enumerate(x_columns)))

        result = qs.group({
            '$group': {
                '_id': dict(map(lambda x: (x['name'], x['lookup']), x_lookups)),
                'aggregation': y_func
            }
        }, {
            '$sort': dict(map(lambda x: (
                '_id.{}'.format(group_name(x[0])),
                pymongo.ASCENDING
            ), enumerate(x_columns)))
        })

        result = list(map(lambda x: {**x['_id'], 'y_func': x['aggregation']}, result))
        return result
    else:
        x_columns = list(map(lambda x: getattr(Model, x), value['x_columns']))
        y_column = getattr(Model, value['y_column'])
        y_func = get_sql_aggregate_func_by_name(value['y_func'], y_column)

        if y_func is None:
            return qs.filter(empty_filter(Model))

        def group_name(i):
            if i == 0 and get_session_engine(qs.session) != 'mssql':
                return 'group'
            else:
                return 'group_{}'.format(i + 1)

        def map_lookup(column, i):
            lookup_name = value['x_lookups'][i] if i < len(value['x_lookups']) else None
            lookup_params = lookup_name.split('_') if lookup_name else []
            lookup_type = lookup_params[0] if len(lookup_params) >= 1 else None
            lookup_param = lookup_params[1] if len(lookup_params) >= 2 else None

            if lookup_type == 'date':
                mapper = inspect_uniform(Model)
                column = mapper.columns.get(column.name)
                if column is not None and not isinstance(column.type, (sqltypes.DateTime, sqltypes.Date)):
                    raise ValidationError('Can\'t apply date functions to non-date field: {}'.format(column.name))

            lookup = get_sql_group_func_lookup(qs.session, lookup_type, lookup_param, column)
            return lookup.label(group_name(i))

        x_lookups = list(map(lambda x: map_lookup(x[1], x[0]), enumerate(x_columns)))
        x_lookup_names = list(map(lambda x: x.name, x_lookups))

        whereclause = qs.whereclause
        qs = qs.session.query(*x_lookups, y_func.label('y_func'))

        if whereclause is not None:
            qs = qs.filter(whereclause)

        if get_session_engine(qs.session) == 'mssql':
            qs = qs.group_by(*x_lookups).order_by(*x_lookup_names)
        else:
            qs = qs.group_by(*x_lookup_names).order_by(*x_lookup_names)

        return qs.limit(1000)


def queryset_search(qs, mapper, search):
    if isinstance(qs, MongoQueryset):
        for index in qs.query.list_indexes():
            if 'text' in index['key'].values():
                return qs.search(search)

        def map_column(column):
            if column.type in [data_types.INTEGER, data_types.FLOAT]:
                try:
                    return column.__eq__(int(search))
                except:
                    pass
            elif column.type in [data_types.BINARY]:
                try:
                    return column.__eq__(ObjectId(search))
                except:
                    pass
            elif column.type in [data_types.CHAR, data_types.TEXT]:
                return column.ilike('%{}%'.format(search))

        operators = list(filter(lambda x: x is not None, map(map_column, mapper.columns)))

        if len(operators) > 1:
            return qs.filter(MongoOperator('or', operators))
        elif len(operators) == 1:
            return qs.filter(operators[0])
        else:
            return qs.filter(mapper.columns['_id'].exists(False))
    else:
        def map_column(column):
            if isinstance(column.type, (sqlalchemy.Integer, sqlalchemy.Numeric)):
                return cast(column, sqlalchemy.String).__eq__(search)
            elif isinstance(column.type, (sqlalchemy.JSON, sqlalchemy.Enum)):
                return cast(column, sqlalchemy.String).ilike('%{}%'.format(search))
            elif isinstance(column.type, sqlalchemy.String):
                return cast(column, sqlalchemy.String).ilike('%{}%'.format(search))

        operators = list(filter(lambda x: x is not None, map(map_column, mapper.columns)))

        return qs.filter(or_(*operators))



from bson import ObjectId
import binascii

from jet_bridge_base.fields.field import Field


class BinaryField(Field):
    def to_internal_value_item(self, value):
        return binascii.unhexlify(value)

    def to_representation_item(self, value):
        if isinstance(value, bytes):
            return binascii.hexlify(value).decode('ascii')
        elif isinstance(value, ObjectId):
            return binascii.hexlify(value.binary).decode('ascii')



from sqlalchemy.exc import SQLAlchemyError

from jet_bridge_base.db_types import apply_session_timezone
from jet_bridge_base.db_types.mongo import MongoSession
from jet_bridge_base.exceptions.not_found import NotFound
from jet_bridge_base.paginators.page_number import PageNumberPagination
from jet_bridge_base.serializers.model_serializer import get_column_data_type
from jet_bridge_base.views.base.api import APIView


class GenericAPIView(APIView):
    serializer_class = None
    filter_class = None
    pagination_class = PageNumberPagination
    _paginator = None
    lookup_url_kwarg = None

    def get_model(self, request):
        raise NotImplementedError

    def get_model_lookup_field(self, request):
        raise NotImplementedError

    def get_queryset(self, request):
        raise NotImplementedError

    def get_object(self, request):
        queryset = self.filter_queryset(request, self.get_queryset(request))
        lookup_url_kwarg = self.lookup_url_kwarg or 'pk'

        if lookup_url_kwarg not in request.path_kwargs:
            raise AssertionError()

        model = self.get_model(request)
        lookup_field = self.get_model_lookup_field(request)
        model_field = getattr(model, lookup_field)

        try:
            field_lookup = getattr(model_field, '__eq__')

            lookup_value = request.path_kwargs[lookup_url_kwarg]
            data_type = get_column_data_type(model_field)
            field = data_type(context={'model_field': model_field})
            lookup_value = field.to_internal_value(lookup_value)

            obj = queryset.filter(field_lookup(lookup_value)).first()
        except SQLAlchemyError:
            queryset.session.rollback()
            raise

        if obj is None:
            raise NotFound

        self.check_object_permissions(request, obj)

        return obj

    def get_filter(self, request, *args, **kwargs):
        filter_class = self.get_filter_class(request)
        if not filter_class:
            return
        kwargs['context'] = self.filter_context()
        return filter_class(*args, **kwargs)

    def get_filter_class(self, request):
        return self.filter_class

    def filter_context(self):
        return {
            # 'request': self.request,
            'handler': self
        }

    def apply_timezone(self, request):
        timezone = request.get_argument('tz', None)
        if timezone is not None:
            if not isinstance(request.session, MongoSession):
                try:
                    apply_session_timezone(request.session, timezone)
                except SQLAlchemyError:
                    request.session.rollback()

    def filter_queryset(self, request, queryset):
        filter_instance = self.get_filter(request)
        if filter_instance:
            queryset = filter_instance.filter_queryset(request, queryset)
        return queryset

    @property
    def paginator(self):
        if not self._paginator:
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class()
        return self._paginator

    def paginate_queryset(self, request, queryset):
        if self.paginator is None:
            return None
        return self.paginator.paginate_queryset(request, queryset, self)

    def get_paginated_response(self, request, data):
        if self.paginator is None:
            raise AssertionError()
        return self.paginator.get_paginated_response(request, data)

    def get_serializer(self, request, *args, **kwargs):
        serializer_class = self.get_serializer_class(request)
        kwargs['context'] = self.get_serializer_context(request)
        return serializer_class(*args, **kwargs)

    def get_serializer_class(self, request):
        return self.serializer_class

    def get_serializer_context(self, request):
        return {
            'request': request,
            'view': self,
            'session': request.session
        }

    def write_error(self, status_code, **kwargs):
        self.session.rollback()
        super(GenericAPIView, self).write_error(status_code, **kwargs)



import re
import time
import graphene
from sqlalchemy.engine import Row
from sqlalchemy.orm import MANYTOONE, ONETOMANY

from jet_bridge_base.db import get_mapped_base, get_engine, get_request_connection
from jet_bridge_base.db_types import desc_uniform, inspect_uniform, get_session_engine, queryset_count_optimized, \
    apply_default_ordering, queryset_search, queryset_group, aliased_uniform
from jet_bridge_base.db_types.sql import sql_load_database_table
from jet_bridge_base.filters import lookups
from jet_bridge_base.filters.filter import EMPTY_VALUES
from jet_bridge_base.filters.filter_for_dbfield import filter_for_column
from jet_bridge_base.models.model_relation_override import ModelRelationOverrideModel
from jet_bridge_base.serializers.model import get_model_serializer
from jet_bridge_base.store import store
from jet_bridge_base.utils.common import get_set_first, any_type_sorter, unique, flatten
from jet_bridge_base.utils.gql import RawScalar
from jet_bridge_base.utils.relations import parse_relationship_direction
from jet_bridge_base.utils.tables import get_table_name


class ModelFiltersType(graphene.InputObjectType):
    pass


class ModelFiltersFieldType(graphene.InputObjectType):
    pass


class ModelFiltersRelationshipType(graphene.InputObjectType):
    pass


class ModelLookupsType(graphene.InputObjectType):
    pass


class ModelLookupsFieldType(graphene.InputObjectType):
    pass


class ModelLookupsRelationshipType(graphene.InputObjectType):
    pass


class ModelSortType(graphene.InputObjectType):
    pass


class ModelAttrsType(graphene.ObjectType):
    pass


class FieldSortType(graphene.InputObjectType):
    descending = graphene.Boolean(required=False)


class PaginationType(graphene.InputObjectType):
    page = graphene.Int()
    offset = graphene.Int()
    limit = graphene.Int()


class SearchType(graphene.InputObjectType):
    query = graphene.String()


class AggregateFuncType(graphene.Enum):
    Count = 'count'
    Sum = 'sum'
    Min = 'min'
    Max = 'max'
    Avg = 'avg'


class AggregateType(graphene.InputObjectType):
    func = AggregateFuncType()
    attr = graphene.String(required=False)


class PaginationResponseType(graphene.ObjectType):
    count = graphene.Int()
    limit = graphene.Int()
    offset = graphene.Int(required=False)
    page = graphene.Int(required=False)
    hasMore = graphene.Boolean(required=False)


def clean_name(name):
    if name == '_meta':
        return '__meta'
    name = re.sub(r'[^_a-zA-Z0-9]', r'_', name)
    name = re.sub(r'^(\d)', r'_\1', name)
    return name


def clean_keys(obj):
    pairs = map(lambda x: [clean_name(x[0]), x[1]], obj.items())
    return dict(pairs)


def get_model_filters_type_relationship_type(self, MappedBase, mapper, relationship, with_relations, depth):
    if not with_relations:
        return

    relationship_filters_type = self.get_model_relationship_filters_type(MappedBase, mapper, relationship, with_relations, depth)
    return relationship_filters_type()


def get_model_filters_type_not_type(self, MappedBase, mapper, with_relations, depth):
    if not with_relations:
        return

    return self.get_model_filters_type(MappedBase, mapper, depth + 1)


def get_model_field_filters_type_relation_type(self, MappedBase, relationship, with_relations, depth):
    if not with_relations:
        return

    relation_mapper = relationship['related_mapper']
    column_filters_type = self.get_model_filters_type(MappedBase, relation_mapper, depth + 1)
    return column_filters_type


def get_model_lookups_type_relation_type(self, MappedBase, mapper, relationship, with_relations, depth):
    if not with_relations:
        return

    relationship_lookups_type = self.get_model_relationship_lookups_type(MappedBase, mapper, relationship, with_relations, depth)
    return relationship_lookups_type()


def get_model_field_lookups_type_relation_type(self, MappedBase, mapper, column_name, with_relations, depth):
    if not with_relations:
        return

    for relationship in self.get_model_relationships(MappedBase, mapper):
        if relationship['direction'] != MANYTOONE or relationship['local_column_name'] != column_name:
            continue

        relation_mapper = relationship['related_mapper']
        lookups_type = self.get_model_lookups_type(MappedBase, relation_mapper, depth + 1)
        return lookups_type()


def get_model_relationship_lookups_type_relation_type(self, MappedBase, relationship, with_relations, depth):
    if not with_relations:
        return

    lookups_type = self.get_model_lookups_type(MappedBase, relationship['related_mapper'], depth + 1)
    return lookups_type()


def apply_dynamic_type(func, *arg, **kwargs):
    class DynamicInstance(graphene.Dynamic):
        def __init__(self, with_schema=False, _creation_counter=None):
            super(DynamicInstance, self).__init__(func, with_schema, _creation_counter)

        def get_type(self, schema=None):
            if schema and self.with_schema:
                return self.type(schema=schema, *arg, **kwargs)
            return self.type(*arg, **kwargs)

    return DynamicInstance()


class GraphQLSchemaGenerator(object):
    def __init__(self):
        self.relationships_by_name = dict()
        self.relationships_by_clean_name = dict()
        self.model_filters_types = dict()
        self.model_filters_field_types = dict()
        self.model_filters_relationship_types = dict()
        self.model_lookups_types = dict()
        self.model_lookups_field_types = dict()
        self.model_lookups_relationship_types = dict()
        self.model_sort_types = dict()

    def get_queryset(self, request, Model, only_columns=None):
        mapper = inspect_uniform(Model)
        pks = mapper.primary_key

        if only_columns:
            only_columns_name = list(map(lambda x: x.name, only_columns))
            missing_pks = list(filter(lambda x: x.name not in only_columns_name, pks))

            if len(missing_pks):
                only_columns = [*missing_pks, *only_columns]

            queryset = request.session.query(*only_columns)
        else:
            queryset = request.session.query(Model)

        mapper = inspect_uniform(Model)
        auto_pk = getattr(mapper.tables[0], '__jet_auto_pk__', False) if len(mapper.tables) else None
        if auto_pk:
            queryset = queryset.filter(mapper.primary_key[0].isnot(None))

        if not auto_pk and get_session_engine(request.session) in ['postgresql', 'mysql']:
            queryset = queryset.group_by(*pks)

        return queryset

    def get_relationships(self, request, MappedBase, draft):
        result = {}
        relationships_overrides = {}

        if store.is_ok():
            connection = get_request_connection(request)

            with store.session() as session:
                overrides = session.query(ModelRelationOverrideModel).filter(
                    ModelRelationOverrideModel.connection_id == connection['id'],
                    draft == draft
                ).all()

                for override in overrides:
                    if override.model not in relationships_overrides:
                        relationships_overrides[override.model] = []
                    relationships_overrides[override.model].append(override)

        for Model in MappedBase.classes:
            model_relationships = {}

            mapper = inspect_uniform(Model)
            name = get_table_name(MappedBase.metadata, mapper.selectable)

            model_relationships_overrides = relationships_overrides.get(name, [])

            for override in model_relationships_overrides:
                direction = parse_relationship_direction(override.direction)
                local_column = getattr(Model, override.local_field, None)

                if local_column is None:
                    continue

                related_name = override.related_model
                related_model = MappedBase.classes.get(related_name)

                if not related_model and '.' in related_name:
                    schema, table = related_name.split('.', 1)
                    engine = get_engine(request)
                    related_model = sql_load_database_table(engine, MappedBase, schema, table)

                if not related_model:
                    continue

                related_mapper = inspect_uniform(related_model)
                related_column = getattr(related_model, override.related_field, None)

                if related_column is None:
                    continue

                model_relationships[override.name] = {
                    'name': override.name,
                    'direction': direction,
                    'local_column': local_column,
                    'local_column_name': override.local_field,
                    'related_model': related_model,
                    'related_mapper': related_mapper,
                    'related_column': related_column,
                    'related_column_name': override.related_field
                }

            for relationship in mapper.relationships.values():
                local_column = get_set_first(relationship.local_columns)
                relation_column = get_set_first(relationship.remote_side)

                if relationship.direction == MANYTOONE:
                    table = relationship.mapper.tables[0]
                    related_name = get_table_name(MappedBase.metadata, table)
                    related_model = MappedBase.classes.get(related_name)

                    model_relationships[relationship.key] = {
                        'name': relationship.key,
                        'direction': relationship.direction,
                        'local_column': local_column,
                        'local_column_name': local_column.name if local_column is not None else None,
                        'related_model': related_model,
                        'related_mapper': relationship.mapper,
                        'related_column': relation_column,
                        'related_column_name': relation_column.name if relation_column is not None else None
                    }
                elif relationship.direction == ONETOMANY:
                    table = relationship.mapper.tables[0]
                    related_name = get_table_name(MappedBase.metadata, table)
                    related_model = MappedBase.classes.get(related_name)

                    model_relationships[relationship.key] = {
                        'name': relationship.key,
                        'direction': relationship.direction,
                        'local_column': local_column,
                        'local_column_name': local_column.name if local_column is not None else None,
                        'related_model': related_model,
                        'related_mapper': relationship.mapper,
                        'related_column': relation_column,
                        'related_column_name': relation_column.name if relation_column is not None else None
                    }

            result[name] = model_relationships

        return result

    def clean_relationships_by_name(self, relationships):
        def map_model_relations(x):
            return clean_name(x[0]), x[1]

        def map_models(x):
            return x[0], dict(map(lambda r: map_model_relations(r), x[1].items()))

        return dict(map(lambda x: map_models(x), relationships.items()))

    def get_model_columns_by_clean_name(self, MappedBase, mapper):
        table = mapper.tables[0]
        name = get_table_name(MappedBase.metadata, table)
        Model = MappedBase.classes.get(name)
        return dict(map(lambda x: (clean_name(x), getattr(Model, x)), mapper.columns.keys()))

    def get_model_relationships(self, MappedBase, mapper):
        name = get_table_name(MappedBase.metadata, mapper.selectable)
        return self.relationships_by_name.get(name, {}).values()

    def get_model_relationships_by_name(self, MappedBase, mapper):
        name = get_table_name(MappedBase.metadata, mapper.selectable)
        return self.relationships_by_name.get(name, {})

    def get_model_relationships_by_clean_name(self, MappedBase, mapper):
        name = get_table_name(MappedBase.metadata, mapper.selectable)
        return self.relationships_by_clean_name.get(name, {})

    def filter_queryset(self, request, MappedBase, queryset, mapper, filters, parent_relations=None, exclude=False):
        parent_relations = parent_relations or []

        for filters_item in filters:
            filters_item_dict = dict(filters_item)

            for filter_name, filter_lookups in filters_item_dict.items():
                if filter_name == '_not_':
                    queryset = self.filter_queryset(
                        request,
                        MappedBase,
                        queryset,
                        mapper,
                        filter_lookups,
                        parent_relations,
                        exclude=True
                    )
                    continue

                columns_by_clean_name = self.get_model_columns_by_clean_name(MappedBase, mapper)
                column = columns_by_clean_name.get(filter_name)
                filter_relationship = self.get_model_relationships_by_clean_name(MappedBase, mapper).get(filter_name)

                if filter_relationship is not None:
                    filter_lookups_dict = dict(filter_lookups)

                    for lookup_name, lookup_value in filter_lookups_dict.items():
                        if lookup_name == 'relation':
                            relation_mapper = filter_relationship['related_mapper']
                            queryset = self.filter_queryset(
                                request,
                                MappedBase,
                                queryset,
                                relation_mapper,
                                lookup_value,
                                [*parent_relations, filter_relationship],
                                exclude
                            )
                elif column is not None:
                    filter_lookups_dict = dict(filter_lookups)

                    for lookup_name, lookup_value in filter_lookups_dict.items():
                        if lookup_name == 'relation':
                            for relationship in self.get_model_relationships(MappedBase, mapper):
                                if relationship['direction'] != MANYTOONE or relationship['local_column_name'] != column.name:
                                    continue

                                relation_mapper = relationship['related_mapper']
                                queryset = self.filter_queryset(
                                    request,
                                    MappedBase,
                                    queryset,
                                    relation_mapper,
                                    lookup_value,
                                    [*parent_relations, relationship],
                                    exclude
                                )
                                break
                        else:
                            if len(parent_relations):
                                last_related_model = None

                                for relationship in parent_relations:
                                    related_model = aliased_uniform(relationship['related_model'])
                                    related_column = getattr(related_model, relationship['related_column'].name)
                                    local_column = getattr(last_related_model, relationship['local_column'].name) if last_related_model else relationship['local_column']
                                    queryset = queryset.join(
                                        related_model,
                                        related_column == local_column
                                    )
                                    last_related_model = related_model

                                column = getattr(last_related_model, column.name)

                            item = filter_for_column(column)
                            lookup = lookups.by_gql.get(lookup_name)
                            filters_instance = item['filter_class'](
                                name=column.key,
                                column=column,
                                lookup=lookup,
                                exclude=False
                            )

                            lookup_value = filters_instance.clean_value(lookup_value)
                            if lookup_value not in EMPTY_VALUES:
                                criterion = filters_instance.get_lookup_criterion(queryset, lookup_value)
                                criterion = ~criterion if exclude else criterion

                                queryset = queryset.filter(criterion)

        return queryset

    def search_queryset(self, queryset, mapper, search):
        if search is not None:
            query = search['query']
            queryset = queryset_search(queryset, mapper, query)

        return queryset

    def get_models_lookups(self, request, MappedBase, models, Model, mapper, lookups):
        result = []

        for lookup_item in lookups:
            lookup_result = self.get_models_lookup(
                lookup_item,
                request,
                MappedBase,
                models,
                Model,
                mapper
            )
            result.append(lookup_result)

        return result

    def get_models_lookup(self, lookup_item, request, MappedBase, models, Model, mapper):
        result = {}

        lookup_item_dict = dict(lookup_item)

        for lookup_name, lookup_data in lookup_item_dict.items():
            columns_by_clean_name = self.get_model_columns_by_clean_name(MappedBase, mapper)
            column = columns_by_clean_name.get(lookup_name)
            relationship = self.get_model_relationships_by_clean_name(MappedBase, mapper).get(lookup_name)

            if relationship is not None:
                lookup_result = {}
                local_column = relationship['local_column']
                lookup_values = sorted(unique(flatten(map(lambda x: getattr(x, local_column.name), models))), key=any_type_sorter)

                lookup_result['return'] = lookup_data.get('return', False)
                lookup_result['return_list'] = lookup_data.get('returnList', False)
                lookup_result['Model'] = Model
                lookup_result['mapper'] = mapper
                lookup_result['model_values'] = list(map(lambda x: {'instance': x, 'value': getattr(x, local_column.name)}, models))
                lookup_result['source_column'] = local_column.name

                if 'aggregate' in lookup_data:
                    relation_model = relationship['related_model']
                    relation_mapper = relationship['related_mapper']
                    relation_column = relationship['related_column']

                    if 'attr' in lookup_data['aggregate']:
                        aggregate_column_name = lookup_data['aggregate']['attr']
                    else:
                        aggregate_column_name = relation_mapper.primary_key[0].name

                    aggregate_column = getattr(relation_model, aggregate_column_name, None)
                    if aggregate_column is not None:
                        queryset = request.session.query(relation_model).filter(relation_column.in_(lookup_values))
                        groups = queryset_group(relation_model, queryset, {
                            'y_func': lookup_data['aggregate']['func'],
                            'y_column': aggregate_column.name,
                            'x_columns': [relation_column.name],
                            'x_lookups': ['plain']
                        })
                        groups_dict = dict(map(lambda x: (x['group'], x['y_func']), groups))

                        lookup_result['aggregated_values'] = list(map(lambda x: {
                            'instance': x,
                            'value': groups_dict.get(getattr(x, local_column.name), 0)
                        }, models))
                        lookup_result['related_column'] = relation_column.name

                if 'relation' in lookup_data:
                    relation_model = relationship['related_model']
                    relation_mapper = relationship['related_mapper']
                    relation_column = relationship['related_column']

                    related_models = request.session\
                        .query(relation_model)\
                        .filter(relation_column.in_(lookup_values))\
                        .all()
                    related_models = list(related_models)

                    lookup_result['related'] = self.get_models_lookup(
                        lookup_data['relation'],
                        request,
                        MappedBase,
                        related_models,
                        relation_model,
                        relation_mapper
                    )
                    lookup_result['related_column'] = relation_column.name

                result[lookup_name] = lookup_result
            elif column is not None:
                lookup_result = {}
                lookup_values = sorted(unique(flatten(map(lambda x: getattr(x, column.name), models))), key=any_type_sorter)

                lookup_result['return'] = lookup_data.get('return', False)
                lookup_result['return_list'] = lookup_data.get('returnList', False)
                lookup_result['Model'] = Model
                lookup_result['mapper'] = mapper
                lookup_result['model_values'] = list(map(lambda x: {'instance': x, 'value': getattr(x, column.name)}, models))
                lookup_result['source_column'] = column.name

                if 'relation' in lookup_data:
                    for relationship in self.get_model_relationships(MappedBase, mapper):
                        if relationship['direction'] != MANYTOONE or relationship['local_column_name'] != column.name:
                            continue

                        relation_mapper = relationship['related_mapper']
                        relation_model = relationship['related_model']
                        relation_column = relationship['related_column']

                        related_models = request.session\
                            .query(relation_model)\
                            .filter(relation_column.in_(lookup_values))\
                            .all()
                        related_models = list(related_models)

                        lookup_result['related'] = self.get_models_lookup(
                            lookup_data['relation'],
                            request,
                            MappedBase,
                            related_models,
                            relation_model,
                            relation_mapper
                        )
                        lookup_result['related_column'] = relation_column.name
                        break

                result[lookup_name] = lookup_result

        return result

    def filter_lookup_models(self, lookup, instance_predicate=None):
        result = {}

        for lookup_name, lookup_data in lookup.items():
            item_result = {}

            model_values = lookup_data['model_values']

            if instance_predicate:
                model_values = list(filter(lambda x: instance_predicate(x['instance']), model_values))

            values = list(flatten(map(lambda x: x['value'], model_values)))

            if lookup_data['return']:
                if lookup_data['return_list']:
                    item_result['value'] = values
                else:
                    item_result['value'] = values[0] if len(values) else None

            if 'related' in lookup_data:
                item_result['related'] = self.filter_lookup_models(
                    lookup_data['related'],
                    lambda x: getattr(x, lookup_data['related_column'], None) in values
                )

            if 'aggregated_values' in lookup_data:
                model_values = list(filter(
                    lambda x: getattr(x['instance'], lookup_data['source_column'], None) in values,
                    lookup_data['aggregated_values']
                ))
                item_result['aggregated'] = model_values[0]['value'] if len(model_values) else 0

            result[lookup_name] = item_result

        return result

    def map_sort_order_field(self, MappedBase, mapper, name, options):
        descending = options.get('descending', False)

        columns_by_clean_name = self.get_model_columns_by_clean_name(MappedBase, mapper)
        column = columns_by_clean_name.get(name)

        if column is None:
            return

        if descending:
            column = desc_uniform(column)

        return column

    def sort_queryset(self, queryset, MappedBase, mapper, sort):
        total_order_by = []

        for item in sort:
            item_dict = dict(item)

            order_by = map(lambda x: self.map_sort_order_field(MappedBase, mapper, x[0], x[1]), item_dict.items())
            order_by = filter(lambda x: x is not None, order_by)
            total_order_by.extend(order_by)

        queryset = queryset.order_by(*total_order_by)

        table = mapper.tables[0]
        name = get_table_name(MappedBase.metadata, table)
        Model = MappedBase.classes.get(name)
        queryset = apply_default_ordering(Model, queryset)

        return queryset

    def get_pagination_limit(self, pagination):
        default_limit = 20
        limit = pagination.get('limit', default_limit)

        if limit > 1000:
            limit = 1000
        elif limit < 0:
            limit = default_limit

        return limit

    def paginate_queryset(self, queryset, pagination):
        limit = self.get_pagination_limit(pagination)

        if 'offset' in pagination:
            queryset = queryset.offset(pagination['offset'])
        elif 'page' in pagination:
            queryset = queryset.offset((pagination['page'] - 1) * limit)

        queryset = queryset.limit(limit)

        return queryset

    def get_model_filters_type(self, MappedBase, mapper, depth=1):
        with_relations = depth <= 4
        table_name = get_table_name(MappedBase.metadata, mapper.selectable)
        model_name = clean_name(table_name)
        cls_name = 'Model{}FiltersType'.format(model_name)

        if cls_name in self.model_filters_types:
            return graphene.List(self.model_filters_types[cls_name])

        attrs = {}

        for column in mapper.columns:
            column_filters_type = self.get_model_field_filters_type(MappedBase, mapper, column, with_relations, depth)
            attr_name = clean_name(column.name)
            attrs[attr_name] = column_filters_type()

        for relationship in self.get_model_relationships(MappedBase, mapper):
            if relationship['direction'] != ONETOMANY:
                continue

            attr_name = clean_name(relationship['name'])
            attrs[attr_name] = apply_dynamic_type(get_model_filters_type_relationship_type, self, MappedBase, mapper, relationship, with_relations, depth)

        attrs['_not_'] = apply_dynamic_type(get_model_filters_type_not_type, self, MappedBase, mapper, with_relations, depth)

        cls = type(cls_name, (ModelFiltersType,), attrs)
        self.model_filters_types[cls_name] = cls
        return graphene.List(cls)

    def get_model_field_filters_type_relationship(self, MappedBase, mapper, column_name):
        for relationship in self.get_model_relationships(MappedBase, mapper):
            if relationship['direction'] != MANYTOONE or relationship['local_column_name'] != column_name:
                continue

            return relationship

    def get_model_field_filters_type(self, MappedBase, mapper, column, with_relations, depth=1):
        table_name = get_table_name(MappedBase.metadata, mapper.selectable)
        model_name = clean_name(table_name)
        column_name = clean_name(column.name)
        dbfield_filter = filter_for_column(column)
        relationship = self.get_model_field_filters_type_relationship(MappedBase, mapper, column_name) if with_relations else None
        cls_name = 'Model{}Column{}FiltersType'.format(model_name, column_name) if relationship \
            else 'Lookups{}FiltersType'.format(dbfield_filter['lookups_name'])

        if cls_name in self.model_filters_field_types:
            return self.model_filters_field_types[cls_name]

        attrs = {}

        for lookup in dbfield_filter['lookups']:
            gql_lookup = lookups.gql.get(lookup)
            gql_scalar = lookups.gql_scalar.get(lookup, RawScalar())
            attrs[gql_lookup] = gql_scalar

        if relationship:
            attrs['relation'] = apply_dynamic_type(get_model_field_filters_type_relation_type, self, MappedBase, relationship, with_relations, depth)

        cls = type(cls_name, (ModelFiltersFieldType,), attrs)
        self.model_filters_field_types[cls_name] = cls
        return cls

    def get_model_relationship_filters_type(self, MappedBase, mapper, relationship, with_relations, depth=1):
        table_name = get_table_name(MappedBase.metadata, mapper.selectable)
        model_name = clean_name(table_name)
        relationship_key = clean_name(relationship['name'])
        cls_name = 'Model{}Relation{}RelationshipType'.format(model_name, relationship_key)

        if cls_name in self.model_filters_relationship_types:
            return self.model_filters_relationship_types[cls_name]

        attrs = {}

        lookups_type = self.get_model_filters_type(MappedBase, relationship['related_mapper'], depth + 1)
        attrs['relation'] = lookups_type

        cls = type(cls_name, (ModelFiltersRelationshipType,), attrs)
        self.model_filters_relationship_types[cls_name] = cls
        return cls

    def get_model_lookups_type(self, MappedBase, mapper, depth=1):
        with_relations = depth <= 4
        table_name = get_table_name(MappedBase.metadata, mapper.selectable)
        model_name = clean_name(table_name)
        cls_name = 'Model{}LookupsType'.format(model_name)

        if cls_name in self.model_lookups_types:
            return self.model_lookups_types[cls_name]

        attrs = {}

        for column in mapper.columns:
            column_lookups_type = self.get_model_field_lookups_type(MappedBase, mapper, column, with_relations, depth)
            attr_name = clean_name(column.name)
            attrs[attr_name] = column_lookups_type()

        for relationship in self.get_model_relationships(MappedBase, mapper):
            if relationship['direction'] != ONETOMANY:
                continue

            attr_name = clean_name(relationship['name'])
            attrs[attr_name] = apply_dynamic_type(get_model_lookups_type_relation_type, self, MappedBase, mapper, relationship, with_relations, depth)

        cls = type(cls_name, (ModelLookupsType,), attrs)
        self.model_lookups_types[cls_name] = cls
        return cls

    def get_model_field_lookups_type_relationship(self, MappedBase, mapper, column_name):
        for relationship in self.get_model_relationships(MappedBase, mapper):
            if relationship['direction'] != MANYTOONE or relationship['local_column_name'] != column_name:
                continue

            return relationship

    def get_model_field_lookups_type(self, MappedBase, mapper, column, with_relations, depth=1):
        table_name = get_table_name(MappedBase.metadata, mapper.selectable)
        model_name = clean_name(table_name)
        column_name = clean_name(column.name)
        relationship = self.get_model_field_lookups_type_relationship(MappedBase, mapper, column_name) if with_relations else None
        cls_name = 'Model{}Column{}LookupsFieldType'.format(model_name, column_name) if relationship \
            else 'LookupsFieldType'

        if cls_name in self.model_lookups_field_types:
            return self.model_lookups_field_types[cls_name]

        attrs = {
            'return': graphene.Boolean(),
            'returnList': graphene.Boolean()
        }

        if relationship:
            attrs['relation'] = apply_dynamic_type(get_model_field_lookups_type_relation_type, self, MappedBase, mapper, column_name, with_relations, depth)

        cls = type(cls_name, (ModelLookupsFieldType,), attrs)
        self.model_lookups_field_types[cls_name] = cls
        return cls

    def get_model_relationship_lookups_type(self, MappedBase, mapper, relationship, with_relations, depth=1):
        table_name = get_table_name(MappedBase.metadata, mapper.selectable)
        model_name = clean_name(table_name)
        relationship_key = clean_name(relationship['name'])
        cls_name = 'Model{}Relation{}LookupsRelationshipType'.format(model_name, relationship_key)

        if cls_name in self.model_lookups_relationship_types:
            return self.model_lookups_relationship_types[cls_name]

        attrs = {
            'aggregate': AggregateType()
        }

        attrs['relation'] = get_model_relationship_lookups_type_relation_type(self, MappedBase, relationship, with_relations, depth)

        cls = type(cls_name, (ModelLookupsRelationshipType,), attrs)
        self.model_lookups_relationship_types[cls_name] = cls
        return cls

    def get_model_sort_type(self, MappedBase, mapper):
        table_name = get_table_name(MappedBase.metadata, mapper.selectable)
        model_name = clean_name(table_name)
        cls_name = 'Model{}SortType'.format(model_name)

        if cls_name in self.model_sort_types:
            return graphene.List(self.model_sort_types[cls_name])

        attrs = {}

        for column in mapper.columns:
            attr_name = clean_name(column.name)
            attrs[attr_name] = FieldSortType()

        cls = type(cls_name, (ModelSortType,), attrs)
        self.model_sort_types[cls_name] = cls
        return graphene.List(cls)

    def get_model_attrs_type(self, MappedBase, mapper):
        table_name = get_table_name(MappedBase.metadata, mapper.selectable)
        name = clean_name(table_name)
        attrs = {}

        for column in mapper.columns:
            attr_name = clean_name(column.name)
            attrs[attr_name] = RawScalar()

        return type('Model{}RecordAttrsType'.format(name), (ModelAttrsType,), attrs)

    def get_selections(self, info, path):
        i = 0
        current_field = info.field_asts[0]

        for path_item in path:
            for selection in current_field.selection_set.selections:
                if selection.name.value == path_item:
                    if i == len(path) - 1:
                        return selection.selection_set.selections
                    else:
                        current_field = selection
                        break

            i += 1

    def resolve_model_list(self, MappedBase, Model, mapper, info, filters=None, lookups=None, sort=None, pagination=None, search=None):
        try:
            filters = filters or []
            lookups = lookups or []
            sort = sort or []
            pagination = pagination or {}

            request = info.context.get('request')

            field_selections = self.get_selections(info, ['data', 'attrs']) or []
            field_names = list(map(lambda x: x.name.value, field_selections))
            data_selections = self.get_selections(info, ['data']) or []
            data_names = list(map(lambda x: x.name.value, data_selections))
            model_attrs = dict(map(lambda x: [clean_name(x), getattr(Model, x)], dir(Model)))
            only_columns = list(filter(lambda x: x is not None, map(lambda x: model_attrs.get(x), field_names))) \
                if len(field_names) and 'allAttrs' not in data_names else None

            queryset = self.get_queryset(request, Model, only_columns)

            queryset = self.filter_queryset(request, MappedBase, queryset, mapper, filters)
            queryset = self.search_queryset(queryset, mapper, search)
            queryset = self.sort_queryset(queryset, MappedBase, mapper, sort)

            data_query_start = time.time()
            queryset_page = list(self.paginate_queryset(queryset, pagination))
            data_query_end = time.time()

            request.context['graphql_data_query_time'] = round(data_query_end - data_query_start, 3)

            serializer_class = get_model_serializer(Model)
            serializer_context = {**info.context}

            queryset_page_lookups = self.get_models_lookups(request, MappedBase, queryset_page, Model, mapper, lookups)

            def map_queryset_page_item(row):
                if isinstance(row, Row):
                    data = dict(row)
                else:
                    data = dict(map(lambda x: (clean_name(x.name), getattr(row, x.name)), mapper.columns))

                serializer = serializer_class(instance=data, context=serializer_context)
                serialized = serializer.representation_data
                serialized = clean_keys(serialized)

                return {
                    'attrs': serialized,
                    'allAttrs': serialized,
                    'lookups': list(map(
                        lambda x: self.filter_lookup_models(x, lambda instance: instance == row),
                        queryset_page_lookups
                    ))
                }

            result = {
                'data': list(map(map_queryset_page_item, queryset_page))
            }

            pagination_selections = self.get_selections(info, ['pagination']) or []
            pagination_names = list(map(lambda x: x.name.value, pagination_selections))

            if len(pagination_names):
                limit = self.get_pagination_limit(pagination)
                offset = pagination.get('offset')
                page = pagination.get('page')

                result['pagination'] = {
                    'limit': limit,
                    'offset': offset,
                    'page': page
                }

                if 'count' in pagination_names or 'hasMore' in pagination_names:
                    count_query_start = time.time()
                    if offset == 0 and len(queryset_page) < limit:
                        count = len(queryset_page)
                    elif page == 1 and len(queryset_page) < limit:
                        count = len(queryset_page)
                    else:
                        count = queryset_count_optimized(request.session, queryset)
                    result['pagination']['count'] = count
                    count_query_end = time.time()

                    if offset is not None:
                        result['pagination']['hasMore'] = offset + limit < count
                    elif page is not None:
                        result['pagination']['hasMore'] = page * limit < count

                    request.context['graphql_count_query_time'] = round(count_query_end - count_query_start, 3)

            return result
        except Exception as e:
            raise e

    def get_query_type(self, request, draft, before_resolve=None, on_progress_updated=None):
        MappedBase = get_mapped_base(request)

        if len(MappedBase.classes) == 0:
            raise Exception('No tables found')

        query_attrs = {}

        self.relationships_by_name = self.get_relationships(request, MappedBase, draft)
        self.relationships_by_clean_name = self.clean_relationships_by_name(self.relationships_by_name)

        i = 0
        total = len(MappedBase.classes)

        for Model in MappedBase.classes:
            # Wait to allow other threads execution
            time.sleep(0.01)

            mapper = inspect_uniform(Model)
            table = mapper.tables[0]
            name = get_table_name(MappedBase.metadata, table)
            name = clean_name(name)

            if on_progress_updated:
                on_progress_updated(name, i, total)

            FiltersType = self.get_model_filters_type(MappedBase, mapper)
            LookupsType = self.get_model_lookups_type(MappedBase, mapper)
            SortType = self.get_model_sort_type(MappedBase, mapper)
            ModelAttrsType = self.get_model_attrs_type(MappedBase, mapper)
            ModelType = type('Model{}ModelType'.format(name), (graphene.ObjectType,), {
                'attrs': graphene.Field(ModelAttrsType),
                'allAttrs': graphene.Field(RawScalar),
                'lookups': graphene.List(RawScalar)
            })
            ModelListType = type('Model{}ModelListType'.format(name), (graphene.ObjectType,), {
                'data': graphene.List(ModelType),
                'pagination': graphene.Field(PaginationResponseType)
            })

            def create_list_resolver(Model, mapper):
                def resolver(parent, info, filters=None, lookups=None, sort=None, pagination=None, search=None):
                    request = info.context.get('request')

                    if before_resolve is not None:
                        before_resolve(request=request, mapper=mapper)

                    return self.resolve_model_list(
                        MappedBase,
                        Model,
                        mapper,
                        info,
                        filters=filters,
                        lookups=lookups,
                        sort=sort,
                        pagination=pagination,
                        search=search
                    )
                return resolver

            query_attrs[name] = graphene.Field(
                ModelListType,
                filters=FiltersType,
                lookups=graphene.List(LookupsType),
                sort=SortType,
                pagination=PaginationType(),
                search=SearchType()
            )
            query_attrs['resolve_{}'.format(name)] = create_list_resolver(Model, mapper)

            i += 1

        if on_progress_updated:
            on_progress_updated(None, total, total)

        return type('Query', (graphene.ObjectType,), query_attrs)

    def get_schema(self, request, draft, before_resolve=None, on_progress_updated=None):
        Query = self.get_query_type(request, draft, before_resolve, on_progress_updated)
        return graphene.Schema(query=Query, auto_camelcase=False)



import inspect


def issubclass_safe(x, A_tuple):
    return inspect.isclass(x) and issubclass(x, A_tuple)


def is_instance_or_subclass(x, A_tuple):
    return isinstance(x, A_tuple) or issubclass_safe(x, A_tuple)



import time

from jet_bridge_base import settings
from jet_bridge_base.logger import logger
from jet_bridge_base.utils.conf import get_connection_only_predicate
from jet_bridge_base.utils.process import get_memory_usage, get_memory_usage_human

from .mongo_session import MongoSession
from .mongo_metadata_file import mongo_load_metadata_file, mongo_dump_metadata_file
from .mongo_reflect import reflect_mongodb
from .mongo_base import MongoBase
from .mongo_engine import MongoEngine


def mongodb_init_database_connection(conf, tunnel, id_short, connection_name, schema, pending_connection):
    engine = MongoEngine()
    pending_connection['engine'] = engine

    logger.info('[{}] Connecting to database "{}"...'.format(id_short, connection_name))

    connect_start = time.time()

    database_url = conf.get('url')
    database_name = conf.get('name')

    engine.connect(database_url)
    db = engine.get_db(database_name)
    Session = lambda: MongoSession(db)

    connect_end = time.time()
    connect_time = round(connect_end - connect_start, 3)

    default_timezone = None
    # default_timezone = fetch_default_timezone(session)
    # if default_timezone is not None:
    #     logger.info('[{}] Default timezone detected: "{}"'.format(id_short, default_timezone))
    # else:
    #     logger.info('[{}] Failed to detect default timezone'.format(id_short))

    metadata_dump = mongo_load_metadata_file(conf)

    if metadata_dump:
        metadata = metadata_dump['metadata']

        reflect_time = None
        reflect_memory_usage_approx = None

        logger.info('[{}] Loaded schema cache for "{}"'.format(id_short, connection_name))
    else:
        logger.info('[{}] Getting schema for "{}"...'.format(id_short, connection_name))

        reflect_start_time = time.time()
        reflect_start_memory_usage = get_memory_usage()

        only = get_connection_only_predicate(conf)
        metadata = reflect_mongodb(
            id_short,
            db,
            only=only,
            pending_connection=pending_connection,
            max_read_records=settings.DATABASE_REFLECT_MAX_RECORDS or 1000000
        )

        reflect_end_time = time.time()
        reflect_end_memory_usage = get_memory_usage()
        reflect_time = round(reflect_end_time - reflect_start_time, 3)
        reflect_memory_usage_approx = reflect_end_memory_usage - reflect_start_memory_usage

        mongo_dump_metadata_file(conf, metadata)

    logger.info('[{}] Connected to "{}" (Mem:{})'.format(id_short, connection_name, get_memory_usage_human()))

    MappedBase = MongoBase(metadata)

    result = {
        'engine': engine,
        'Session': Session,
        'MappedBase': MappedBase,
        'default_timezone': default_timezone,
        'connect_time': connect_time,
        'reflect_time': reflect_time,
        'reflect_memory_usage_approx': reflect_memory_usage_approx,
        'reflect_metadata_dump': metadata_dump['file_path'] if metadata_dump else None
    }

    return result


def mongo_load_mapped_base(MappedBase, clear=False):
    raise NotImplementedError



from jet_bridge_base.utils.track_database import track_database_async
from jet_bridge_base.utils.track_model import track_model_async
from sqlalchemy.exc import SQLAlchemyError

from jet_bridge_base import status
from jet_bridge_base.configuration import configuration
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.utils.exceptions import validation_error_from_database_error


class DestroyAPIViewMixin(object):

    def destroy(self, request, *args, **kwargs):
        track_database_async(request)

        self.apply_timezone(request)
        instance = self.get_object(request)
        self.perform_destroy(request, instance)

        serializer = self.get_serializer(request, instance=instance)
        representation_data = serializer.representation_data
        track_model_async(request, kwargs.get('model'), 'delete', kwargs.get('pk'), representation_data)

        return JSONResponse(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, request, instance):
        configuration.on_model_pre_delete(request.path_kwargs['model'], instance)
        Model = self.get_model(request)
        request.session.delete(instance)

        try:
            request.session.commit()
        except SQLAlchemyError as e:
            request.session.rollback()
            raise validation_error_from_database_error(e, Model)

        configuration.on_model_post_delete(request.path_kwargs['model'], instance)



import json
import os
import re

from jet_bridge_base import settings
from jet_bridge_base.utils.common import merge
from jet_bridge_base.utils.crypt import get_sha256_hash
from jet_bridge_base.utils.text import clean_alphanumeric


def get_settings_conf():
    return {
        'engine': settings.DATABASE_ENGINE,
        'url': settings.DATABASE_URL,
        'host': settings.DATABASE_HOST,
        'port': settings.DATABASE_PORT,
        'name': settings.DATABASE_NAME,
        'user': settings.DATABASE_USER,
        'password': settings.DATABASE_PASSWORD,
        'extra': settings.DATABASE_EXTRA,
        'connections': settings.DATABASE_CONNECTIONS,
        'connections_overflow': settings.DATABASE_CONNECTIONS_OVERFLOW,
        'only': settings.DATABASE_ONLY,
        'except': settings.DATABASE_EXCEPT,
        'schema': settings.DATABASE_SCHEMA,
        'ssl_ca': settings.DATABASE_SSL_CA,
        'ssl_cert': settings.DATABASE_SSL_CERT,
        'ssl_key': settings.DATABASE_SSL_KEY,
        'ssh_host': settings.DATABASE_SSH_HOST,
        'ssh_port': settings.DATABASE_SSH_PORT,
        'ssh_user': settings.DATABASE_SSH_USER,
        'ssh_private_key': settings.DATABASE_SSH_PRIVATE_KEY,
        'project': settings.PROJECT,
        'token': settings.TOKEN
    }


def get_request_conf(request):
    bridge_settings = request.get_bridge_settings()

    if not bridge_settings:
        return

    return {
        'engine': bridge_settings.get('database_engine'),
        'url': bridge_settings.get('database_url'),
        'host': bridge_settings.get('database_host'),
        'port': bridge_settings.get('database_port'),
        'name': bridge_settings.get('database_name'),
        'user': bridge_settings.get('database_user'),
        'password': bridge_settings.get('database_password'),
        'extra': bridge_settings.get('database_extra'),
        'connections': bridge_settings.get('database_connections', settings.DATABASE_CONNECTIONS),
        'connections_overflow': bridge_settings.get('database_connections_overflow', settings.DATABASE_CONNECTIONS_OVERFLOW),
        'only': bridge_settings.get('database_only'),
        'except': bridge_settings.get('database_except'),
        'schema': bridge_settings.get('database_schema'),
        'ssl_ca': bridge_settings.get('database_ssl_ca'),
        'ssl_cert': bridge_settings.get('database_ssl_cert'),
        'ssl_key': bridge_settings.get('database_ssl_key'),
        'ssh_host': bridge_settings.get('database_ssh_host'),
        'ssh_port': bridge_settings.get('database_ssh_port'),
        'ssh_user': bridge_settings.get('database_ssh_user'),
        'ssh_private_key': bridge_settings.get('database_ssh_private_key'),
        'project': bridge_settings.get('project'),
        'token': bridge_settings.get('token'),
    }


def get_conf(request):
    request_conf = get_request_conf(request)

    if request_conf:
        return request_conf
    else:
        return get_settings_conf()


def get_connection_id(conf):
    return get_sha256_hash(json.dumps([
        conf.get('engine'),
        conf.get('url'),
        conf.get('host'),
        conf.get('port'),
        conf.get('name'),
        conf.get('user'),
        conf.get('password'),
        conf.get('schema'),
        conf.get('ssl_ca'),
        conf.get('ssl_cert'),
        conf.get('ssl_key'),
        conf.get('ssh_host'),
        conf.get('ssh_port'),
        conf.get('ssh_user'),
        conf.get('ssh_private_key')
    ]))


def get_connection_meta_params_id(conf):
    return get_sha256_hash(json.dumps([
        conf.get('only'),
        conf.get('except'),
        conf.get('extra')
    ]))


def get_connection_params_id(conf):
    return json.dumps([
        conf.get('only'),
        conf.get('except'),
        conf.get('extra'),
        conf.get('connections'),
        conf.get('connections_overflow')
    ])


def is_tunnel_connection(conf):
    return all(map(lambda x: conf.get(x), ['ssh_host', 'ssh_port', 'ssh_user', 'ssh_private_key']))


def get_connection_schema(conf):
    schema = conf.get('schema') if conf.get('schema') and conf.get('schema') != '' else None

    if not schema and str(conf.get('engine', '')).startswith('mssql'):
        schema = 'dbo'

    return schema


def clean_connection_url(url):
    if not isinstance(url, str):
        return url
    return re.sub(r'//([^:]+):[^@/]+@', r'//\1:********@', url)


def get_connection_name(conf, schema):
    if conf.get('engine') == 'mongo':
        url = str(conf.get('url'))

        connection_name = [url]

        if not url.endswith('/'):
            connection_name.append('/')

        connection_name.append(conf.get('name') or '')

        return ''.join(connection_name)
    else:
        from jet_bridge_base.db_types.sql import sql_build_engine_url

        password_token = '__JET_DB_PASS__'
        log_conf = merge(merge({}, conf), {'password': password_token})

        connection_name = sql_build_engine_url(log_conf)
        if connection_name:
            connection_name = connection_name.replace(password_token, '********')
        if schema:
            connection_name += ':{}'.format(schema)
        if is_tunnel_connection(conf):
            connection_name += ' (via {}@{}:{})'.format(conf.get('ssh_user'), conf.get('ssh_host'), conf.get('ssh_port'))

        return connection_name


def get_connection_short_name_parts(conf):
    result = []

    if conf.get('url'):
        result.append(clean_connection_url(str(conf.get('url'))))
    else:
        if conf.get('engine'):
            result.append(str(conf.get('engine')))

        if conf.get('host'):
            result.append(str(conf.get('host')))

        if conf.get('port'):
            result.append(str(conf.get('port')))

    if conf.get('name'):
        result.append(str(conf.get('name')))

    return result


def get_connection_only_predicate(conf):
    def only(table):
        if conf.get('only') is not None and table not in conf.get('only'):
            return False
        if conf.get('except') is not None and table in conf.get('except'):
            return False
        return True
    return only


def get_metadata_file_path(conf):
    short_name = '_'.join(map(lambda x: clean_alphanumeric(x), get_connection_short_name_parts(conf)))
    id_hash = get_connection_id(conf)
    params_id_hash = get_connection_meta_params_id(conf)[:8]
    engine_length = len(str(conf.get('engine'))) + 1 if conf.get('engine') else 0
    file_name = '{}_{}_{}.dump'.format(short_name[:(50 + engine_length)], id_hash, params_id_hash)

    return os.path.join(settings.CACHE_METADATA_PATH, file_name)



from jet_bridge_base.db_types import inspect_uniform, queryset_search
from jet_bridge_base.filters.char_filter import CharFilter
from jet_bridge_base.filters.filter import EMPTY_VALUES


def get_model_search_filter(Model):
    mapper = inspect_uniform(Model)

    class ModelSearchFilter(CharFilter):
        def filter(self, qs, value):
            if value in EMPTY_VALUES:
                return qs

            value = self.clean_value(value)
            return queryset_search(qs, mapper, value)

    return ModelSearchFilter



from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.optional_json import OptionalJSONResponse
from jet_bridge_base.serializers.message import MessageSerializer
from jet_bridge_base.views.base.api import APIView


class MessageView(APIView):
    permission_classes = (HasProjectPermissions,)

    def post(self, request, *args, **kwargs):
        serializer = MessageSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        result = serializer.save()

        return OptionalJSONResponse(result)



from pymongo import MongoClient


class MongoEngine(object):
    client = None

    def connect(self, url):
        self.url = url
        self.client = MongoClient(url)

    def get_db(self, name):
        return self.client[name]

    def dispose(self):
        if self.client:
            self.client.close()
            self.client = None

    def __repr__(self):
        return self.url



from sqlalchemy import inspect, func
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import load_only

from jet_bridge_base.db_types import apply_default_ordering, queryset_count_optimized, get_queryset_order_by


def get_row_number(Model, queryset, instance):
    mapper = inspect(Model)
    pk = mapper.primary_key[0].name
    ordering = get_queryset_order_by(queryset)

    subqquery = queryset.with_entities(
        mapper.primary_key[0].label('__inner__pk'),
        func.row_number().over(order_by=ordering).label('__inner__row')
    ).subquery()

    try:
        rest = queryset.session.query(subqquery.c.__inner__row).filter(subqquery.c.__inner__pk == getattr(instance, pk))
        return rest.scalar()
    except SQLAlchemyError:
        queryset.session.rollback()
        raise


def get_row_siblings(Model, queryset, row_number):
    mapper = inspect(Model)
    pk = mapper.primary_key[0].name

    has_prev = row_number > 1
    offset = row_number - 2 if has_prev else row_number - 1
    limit = 3 if has_prev else 2

    try:
        rows = queryset.options(load_only(pk)).limit(limit).offset(offset).all()
    except SQLAlchemyError:
        queryset.session.rollback()
        raise

    if has_prev:
        next_index = 2
    else:
        next_index = 1

    if next_index >= len(rows):
        next_index = None

    if has_prev:
        prev_index = 0
    else:
        prev_index = None

    def map_row(row):
        return dict(((pk, getattr(row, pk)),))

    return {
        'prev': map_row(rows[prev_index]) if prev_index is not None else None,
        'next': map_row(rows[next_index]) if next_index is not None else None
    }


def get_model_siblings(request, Model, instance, queryset):
    count = queryset_count_optimized(request.session, queryset)

    if count > 10000:
        return {}

    queryset = apply_default_ordering(Model, queryset)
    row_number = get_row_number(Model, queryset, instance)

    if not row_number:
        return {}

    return get_row_siblings(Model, queryset, row_number)



class MongoRecordMeta(object):
    def __init__(self, table_name):
        self.table_name = table_name


class MongoRecord(object):
    create_pending = False
    delete_pending = False

    def __init__(self, table_name, **kwargs):
        object.__setattr__(self, 'meta', MongoRecordMeta(table_name))
        object.__setattr__(self, 'data', dict(kwargs))
        object.__setattr__(self, 'update_pending', set())

    def __getattr__(self, name):
        return self.get(name)

    def get(self, name, default=None):
        return self.data.get(name, default)

    def __setattr__(self, name, value):
        self.set(name, value)

    def set(self, name, value):
        self.data[name] = value
        self.mark_update(name)

    def get_data(self):
        return object.__getattribute__(self, 'data')

    def get_meta(self):
        return object.__getattribute__(self, 'meta')

    def is_create_pending(self):
        return object.__getattribute__(self, 'create_pending')

    def mark_create(self):
        object.__setattr__(self, 'create_pending', True)

    def get_update_pending(self):
        return object.__getattribute__(self, 'update_pending')

    def mark_update(self, name):
        object.__getattribute__(self, 'update_pending').add(name)

    def is_delete_pending(self):
        return object.__getattribute__(self, 'delete_pending')

    def mark_delete(self):
        object.__setattr__(self, 'delete_pending', True)

    def clear_pending(self):
        object.__setattr__(self, 'create_pending', False)
        object.__getattribute__(self, 'update_pending').clear()
        object.__setattr__(self, 'delete_pending', False)

    def __repr__(self):
        return '_id({})'.format(self.get('_id'))



import webbrowser

from requests import RequestException

from jet_bridge_base import settings
from jet_bridge_base.utils.backend import is_resource_token_activated
from jet_bridge_base.logger import logger


def check_token_command(api_url):
    try:
        if not is_resource_token_activated(settings.PROJECT, settings.TOKEN):
            logger.warning('[!] Your resource token is not activated')
            logger.warning('[!] Project: {}'.format(settings.PROJECT))
            logger.warning('[!] Token: {}'.format(settings.TOKEN))

            if settings.DATABASE_ENGINE != 'none' and settings.AUTO_OPEN_REGISTER and api_url.startswith('http'):
                register_url = '{}register/'.format(api_url)

                if settings.TOKEN:
                    register_url += '?token={}'.format(settings.TOKEN)

                if webbrowser.open(register_url):
                    logger.warning('[!] Activation page was opened in your browser - {}'.format(register_url))
            else:
                register_url = '{}register/'.format(api_url)
                logger.warning('[!] Go to {} to activate it'.format(register_url))
    except RequestException:
        logger.error('[!] Can\'t connect to Jet Admin API')
        logger.error('[!] Token verification failed')



import json

from social_core.actions import do_complete

from jet_bridge_base import settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.external_auth.mixin import ExternalAuthMixin
from jet_bridge_base.responses.template import TemplateResponse
from jet_bridge_base.views.base.api import BaseAPIView
from jet_bridge_base.views.external_auth.login import REDIRECT_URI_KEY, AUTH_URI_KEY, PROJECT_KEY


class ExternalAuthCompleteView(ExternalAuthMixin, BaseAPIView):

    def get(self, request, *args, **kwargs):
        backend = kwargs.get('app')
        return self._complete(request, backend)

    def post(self, request, *args, **kwargs):
        backend = kwargs.get('app')
        return self._complete(request, backend)

    def _complete(self, request, app):
        self.init_auth(request, app)

        # Hack for passing SSO
        setattr(self.backend, 'sso', app)
        result = do_complete(
            self.backend,
            login=lambda: None
        )

        success = result and result.get('auth')
        auth_uri = configuration.session_get(request, AUTH_URI_KEY, '/api/')
        project = configuration.session_get(request, PROJECT_KEY)
        redirect_uri = configuration.session_get(request, REDIRECT_URI_KEY)

        data = {
            'sso': app,
            'token': settings.TOKEN,
            'project': project,
            'redirect_uri': redirect_uri
        }

        if success:
            data['result'] = True
            data['username'] = result['details'].get('username')
            data['email'] = result['details'].get('email')
            data['first_name'] = result['details'].get('first_name')
            data['last_name'] = result['details'].get('last_name')
            data['full_name'] = result['details'].get('fullname')
        else:
            data['result'] = False
            data['error'] = 'Authentication failed'

        return TemplateResponse('external_auth_complete.html', status=200, data={
            'context': json.dumps({
                'url': auth_uri,
                'data': data
            })
        })



import sqlalchemy

from jet_bridge_base.db import get_mapped_base
from jet_bridge_base.db_types import inspect_uniform
from jet_bridge_base.filters.char_filter import CharFilter
from jet_bridge_base.filters.filter import EMPTY_VALUES


def filter_search_field(field):
    allowed_fields = [
        sqlalchemy.String,
        sqlalchemy.JSON,
    ]

    return isinstance(field.type, tuple(allowed_fields))


def get_model_relation_filter(request, Model):
    mapper = inspect_uniform(Model)
    MappedBase = get_mapped_base(request)

    class ModelRelationFilter(CharFilter):
        def filter(self, qs, value):
            if value in EMPTY_VALUES:
                return qs

            current_table = mapper.tables[0]
            path = list(map(lambda x: x.split('.'), value.split('|')))
            path_len = len(path)

            for i in range(path_len):
                item = path[i]
                last = i == path_len - 1

                if not last:
                    current_table_column = current_table.columns[item[0]]

                    related_table = MappedBase.metadata.tables[item[1]]
                    related_table_column = related_table.columns[item[2]]

                    qs = qs.join(related_table, current_table_column == related_table_column)
                    current_table = related_table
                else:
                    current_table_column = current_table.columns[item[0]]
                    value = item[1].split(',')
                    qs = qs.filter(current_table_column.in_(value))

            return qs

    return ModelRelationFilter



import json
import os
import tempfile
import time
from json import JSONDecodeError

from jet_bridge_base.exceptions.request_error import RequestError
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.utils.crypt import get_sha256_hash
from jet_bridge_base.utils.process import get_memory_usage
from six import string_types

from jet_bridge_base import settings
from jet_bridge_base.exceptions.missing_argument_error import MissingArgumentError

_ARG_DEFAULT = object()


class Request(object):

    session = None
    bridge_settings = None
    project = None
    environment = None
    resource_token = None
    context = {}

    track_start_time = None
    track_start_memory_usage = None

    def __init__(
            self,
            method=None,
            protocol=None,
            host=None,
            path=None,
            path_kwargs=None,
            uri=None,
            query_arguments=None,
            headers=None,
            body=None,
            body_arguments=None,
            files=None,
            original_request=None,
            original_handler=None,
            action=None
    ):
        self.method = method
        self.protocol = protocol
        self.host = host
        self.path = path
        self.path_kwargs = path_kwargs
        self.uri = uri
        self.query_arguments = query_arguments or {}
        self.headers = headers or {}
        self.body = body
        self.body_arguments = body_arguments or {}
        self.files = files or {}
        self.original_request = original_request
        self.original_handler = original_handler
        self.action = action

        content_type = self.headers.get('CONTENT_TYPE', '')

        if content_type.startswith('application/json'):
            data = self.body

            if not isinstance(data, string_types):
                data = data.decode('utf-8', 'surrogatepass')
            try:
                self.data = json.loads(data) if data else {}
            except ValueError as e:
                raise RequestError(self, 'Incorrect JSON body: {}'.format(e))
        else:
            def body_argument_value(value):
                if isinstance(value, bytes):
                    try:
                        value = value.decode('utf-8')
                    except:
                        pass
                return value

            def map_item(item):
                key, values = item
                values = list(map(body_argument_value, values))
                if len(values) > 1:
                    return (key, values)
                elif len(values) == 1:
                    return (key, values[0])
                else:
                    return (key, None)

            tuples = list(map(map_item, self.body_arguments.items()))
            self.data = dict(tuples)

    def full_url(self):
        return self.protocol + "://" + self.host + self.uri

    def get_argument(self, name, default=_ARG_DEFAULT, strip=True):
        return self._get_argument(name, default, self.query_arguments, strip)

    def get_arguments(self, name, strip=True):
        return self._get_arguments(name, self.query_arguments, strip)

    def get_argument_safe(self, name, default=_ARG_DEFAULT):
        values = self.get_arguments(name)

        if len(values) == 0:
            value = default
        elif len(values) == 1:
            value = values[0]
        else:
            value = values

        return value

    def get_body_argument(self, name, default=_ARG_DEFAULT, strip=True):
        return self._get_argument(name, default, self.body_arguments, strip)

    def get_body_arguments(self, name, strip=True):
        return self._get_arguments(name, self.body_arguments, strip)

    def get_ip(self):
        return self.headers.get('X_REAL_IP')

    def get_stick_session(self):
        return self.headers.get('X_STICK_SESSION')

    def _get_argument(self, name, default, source, strip=True):
        args = self._get_arguments(name, source, strip=strip)
        if not args:
            if default is _ARG_DEFAULT:
                raise MissingArgumentError(name)
            return default
        return args[-1]

    def _get_arguments(self, name, source, strip=True):
        values = []
        for v in source.get(name, []):
            if isinstance(v, bytes):
                v = v.decode('utf-8')
            # v = self.decode_argument(v, name=name)
            # if isinstance(v, unicode_type):
            #     # Get rid of any weird control chars (unless decoding gave
            #     # us bytes, in which case leave it alone)
            #     v = RequestHandler._remove_control_chars_regex.sub(" ", v)
            if strip:
                v = v.strip()
            values.append(v)
        return values

    def get_bridge_settings(self):
        if self.bridge_settings:
            return self.bridge_settings

        bridge_settings_encoded = self.headers.get('X_BRIDGE_SETTINGS')

        if not bridge_settings_encoded:
            return

        from jet_bridge_base.utils.crypt import decrypt

        try:
            secret_key = settings.TOKEN.replace('-', '').lower()
            decrypted = decrypt(bridge_settings_encoded, secret_key)

            self.bridge_settings = {
                **json.loads(decrypted),
                'raw': bridge_settings_encoded
            }
        except Exception:
            return

        database_ssl_ca = self.bridge_settings.get('database_ssl_ca')
        database_ssl_cert = self.bridge_settings.get('database_ssl_cert')
        database_ssl_key = self.bridge_settings.get('database_ssl_key')

        temp_dir = os.path.join(tempfile.gettempdir(), 'ssl')

        try:
            os.makedirs(temp_dir)
        except OSError:
            pass

        if database_ssl_ca:
            self.bridge_settings['database_ssl_ca'] = self.save_file(temp_dir, '{}-ca.pem', database_ssl_ca)

        if database_ssl_cert:
            self.bridge_settings['database_ssl_cert'] = self.save_file(temp_dir, '{}-cert.pem', database_ssl_cert)

        if database_ssl_key:
            self.bridge_settings['database_ssl_key'] = self.save_file(temp_dir, '{}-key.pem', database_ssl_key)

        return self.bridge_settings

    def save_file(self, dir_path, name, content):
        file_hash = get_sha256_hash(content)
        file_path = os.path.join(dir_path, name.format(file_hash))

        with open(file_path, 'w') as f:
            f.write(content)

        return f.name

    def start_track(self):
        self.track_start_time = time.time()
        self.track_start_memory_usage = get_memory_usage()

    def get_track_time(self):
        if self.track_start_time is None:
            return
        return round(time.time() - self.track_start_time, 3)

    def get_track_memory_usage(self):
        if self.track_start_memory_usage is None:
            return
        return get_memory_usage() - self.track_start_memory_usage




GET_ACTION_LIST = 'get_action_list'
EXECUTE_ACTION = 'execute_action'
GET_FIELD_OPTIONS = 'get_field_options'
GET_ELEMENT_STATUS = 'get_element_status'

message_handlers = {}


def add_handler(message_name, func):
    message_handlers[message_name] = func


def get_handler(message_name):
    return message_handlers.get(message_name)



from jet_bridge_base.fields.raw import RawField
from jet_bridge_base.fields.char import CharField
from jet_bridge_base.fields.boolean import BooleanField
from jet_bridge_base.fields.integer import IntegerField
from jet_bridge_base.fields.float import FloatField
from jet_bridge_base.fields.datetime import DateTimeField
from jet_bridge_base.fields.json import JSONField
from jet_bridge_base.fields.wkt import WKTField
from jet_bridge_base.fields.array import ArrayField
from jet_bridge_base.fields.binary import BinaryField



from jet_bridge_base import fields
from jet_bridge_base.db import get_mapped_base, reload_request_graphql_schema, get_request_connection
from jet_bridge_base.db_types import inspect_uniform
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.models.model_relation_override import ModelRelationOverrideModel
from jet_bridge_base.serializers.serializer import Serializer
from jet_bridge_base.store import store
from jet_bridge_base.logger import logger


class ModelDescriptionRelationOverrideSerializer(Serializer):
    direction = fields.CharField()
    local_field = fields.CharField()
    related_model = fields.CharField()
    related_field = fields.CharField()


class ModelDescriptionRelationOverridesSerializer(Serializer):
    model = fields.CharField()
    relations = ModelDescriptionRelationOverrideSerializer(many=True)

    def get_model(self, request, name):
        MappedBase = get_mapped_base(request)
        return MappedBase.classes.get(name)

    def generate_many_to_one_name(self, mapper, local_field, related_model, related_field):
        name = '__'.join([local_field, 'to', related_model, related_field])

        if name in mapper.columns:
            name = name + '_relation'
            logger.warning('Already detected column name, using {}'.format(name))

        return name

    def generate_one_to_many_name(self, mapper, local_field, related_model, related_field):
        name = '__'.join([related_model, related_field, 'to', local_field])

        if name in mapper.columns:
            name = name + '_relation'
            logger.warning('Already detected column name, using {}'.format(name))

        return name

    def validate(self, attrs):
        request = self.context.get('request')

        Model = self.get_model(request, attrs['model'])
        if Model is None:
            raise ValidationError('Unknown relation override model: {}'.format(attrs['model']))

        mapper = inspect_uniform(Model)

        for item in attrs['relations']:
            if item['direction'] == 'MANYTOONE':
                item['name'] = self.generate_many_to_one_name(mapper, item['local_field'], item['related_model'], item['related_field'])
            elif item['direction'] == 'ONETOMANY':
                item['name'] = self.generate_one_to_many_name(mapper, item['local_field'], item['related_model'], item['related_field'])
            else:
                raise ValidationError('Unknown relation direction: {}'.format(item['direction']))

        return attrs

    def save(self):
        request = self.context.get('request')
        connection = get_request_connection(request)
        draft = bool(request.get_argument('draft', False))

        with store.session() as session:
            with session.begin():
                for item in self.validated_data:
                    set_overrides = sorted(item['relations'], key=lambda x: x['name'])

                    existing_overrides = session.query(ModelRelationOverrideModel).filter(
                        ModelRelationOverrideModel.connection_id == connection['id'],
                        ModelRelationOverrideModel.model == item['model'],
                        draft == draft
                    ).order_by(ModelRelationOverrideModel.name).all()
                    existing_overrides = list(existing_overrides)

                    for i, override in enumerate(set_overrides):
                        existing_override = existing_overrides[i] if i < len(existing_overrides) else None

                        if existing_override:
                            existing_override.name = override.get('name')
                            existing_override.direction = override.get('direction')
                            existing_override.local_field = override.get('local_field')
                            existing_override.related_model = override.get('related_model')
                            existing_override.related_field = override.get('related_field')
                        else:
                            session.add(ModelRelationOverrideModel(
                                connection_id=connection['id'],
                                model=item['model'],
                                draft=draft,
                                name=override.get('name'),
                                direction=override.get('direction'),
                                local_field=override.get('local_field'),
                                related_model=override.get('related_model'),
                                related_field=override.get('related_field')
                            ))

                    delete_overrides = existing_overrides[len(item['relations']):]
                    for override in delete_overrides:
                        session.delete(override)

        reload_request_graphql_schema(request, draft)



from .mongo import mongodb_init_database_connection, MongoBase, mongo_load_mapped_base
from .sql import sql_init_database_connection, sql_load_mapped_base


def init_database_connection(conf, tunnel, id_short, connection_name, schema, pending_connection):
    if conf.get('engine') == 'mongo':
        return mongodb_init_database_connection(
            conf,
            tunnel,
            id_short,
            connection_name,
            schema,
            pending_connection
        )
    else:
        return sql_init_database_connection(
            conf,
            tunnel,
            id_short,
            connection_name,
            schema,
            pending_connection
        )


def load_mapped_base(MappedBase, clear=False):
    if isinstance(MappedBase, MongoBase):
        mongo_load_mapped_base(MappedBase, clear)
    else:
        sql_load_mapped_base(MappedBase, clear)



try:
    from collections.abc import Mapping, Iterable
    from collections import OrderedDict
except ImportError:
    from collections import OrderedDict, Mapping, Iterable

import six

from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.fields.field import Field, empty


class SerializerMetaclass(type):

    @classmethod
    def _get_declared_fields(cls, bases, attrs):
        fields = [(field_name, attrs.pop(field_name))
                  for field_name, obj in list(attrs.items())
                  if isinstance(obj, Field)]
        fields.sort(key=lambda x: x[1].creation_counter)

        for base in reversed(bases):
            if hasattr(base, '_declared_fields'):
                fields = [
                    (field_name, obj) for field_name, obj
                    in base._declared_fields.items()
                    if field_name not in attrs
                ] + fields

        return OrderedDict(fields)

    def __new__(cls, name, bases, attrs):
        attrs['_declared_fields'] = cls._get_declared_fields(bases, attrs)
        return super(SerializerMetaclass, cls).__new__(cls, name, bases, attrs)


@six.add_metaclass(SerializerMetaclass)
class Serializer(Field):
    validated_data = None
    fields = []
    errors = None

    def __init__(self, *args, **kwargs):
        self.instance = kwargs.pop('instance', None)
        self.data = kwargs.pop('data', None)
        self.meta = getattr(self, 'Meta', None)
        self.partial = kwargs.pop('partial', False)
        super(Serializer, self).__init__(*args, **kwargs)
        self.update_fields()

    def update_fields(self):
        self.fields = self.get_fields()

    def get_fields(self):
        result = []

        for field_name, field in self._declared_fields.items():
            field.field_name = field_name
            result.append(field)

        return result

    @property
    def readable_fields(self):
        if not self.fields or not isinstance(self.fields, Iterable):
            return []
        return list(filter(lambda x: not x.write_only, self.fields))

    @property
    def writable_fields(self):
        if not self.fields or not isinstance(self.fields, Iterable):
            return []
        return list(filter(lambda x: not x.read_only, self.fields))

    def run_validation(self, value):
        if value is empty:
            if self.required:
                # raise ValidationError('Field is required')
                self.error('required')

        value = self.to_internal_value(value)

        if self.many:
            try:
                value = list(map(lambda x: self.validate(x), value))
                assert value is not None, '.validate() should return the validated data'
            except ValidationError as e:
                raise e
        else:
            try:
                value = self.validate(value)
                assert value is not None, '.validate() should return the validated data'
            except ValidationError as e:
                raise e

        return value

    def is_valid(self, raise_exception=False):
        try:
            self.validated_data = self.run_validation(self.data)
            self.errors = None
        except ValidationError as e:
            self.validated_data = None
            self.errors = e.detail

        if self.errors and raise_exception:
            raise ValidationError(self.errors)

        return not bool(self.errors)

    def to_internal_value_item(self, value):
        result = OrderedDict()
        errors = OrderedDict()

        for field in self.writable_fields:
            field_value = field.get_value(value)

            if field_value is empty:
                if self.partial or not field.required:
                    continue

            validate_method = getattr(self, 'validate_' + field.field_name, None)

            try:
                validated_value = field.run_validation(field_value)
                if validate_method is not None:
                    validated_value = validate_method(validated_value)
                result[field.field_name] = validated_value
            except ValidationError as e:
                errors[field.field_name] = e

        if errors:
            raise ValidationError(errors)

        return result

    def to_representation_item(self, value):
        result = OrderedDict()

        for field in self.readable_fields:
            if isinstance(value, Mapping):
                field_value = value.get(field.field_name, empty)
            else:
                field_value = getattr(value, field.field_name, empty)

            if field_value is empty:
                if not field.required:
                    continue
                else:
                    field_value = None

            result[field.field_name] = field.to_representation(field_value)

        return result

    @property
    def representation_data(self):
        if self.instance is not None:
            return self.to_representation(self.instance)
        elif self.validated_data is not None:
            return self.to_representation(self.validated_data)

    def validate(self, attrs):
        return attrs

    def update(self, instance, validated_data):
        raise NotImplementedError('`update()` must be implemented.')

    def create(self, validated_data):
        raise NotImplementedError('`create()` must be implemented.')

    def save(self, **kwargs):
        if self.errors:
            raise AssertionError('You cannot call `.save()` on a serializer with invalid data.')

        validated_data = dict(
            list(self.validated_data.items()) +
            list(kwargs.items())
        )

        if self.instance is not None:
            self.instance = self.update(self.instance, validated_data)

            if self.instance is None:
                raise AssertionError('`update()` did not return an object instance.')

        else:
            self.instance = self.create(validated_data)

            if self.instance is None:
                raise AssertionError('`create()` did not return an object instance.')

        return self.instance
